#!/usr/bin/python3
"""
    loc_to_geod( loc_coord )
    geod_to_loc(pos_geod )
    gdatum(datum="IGS08")
    xyzell(datum,xyzcoor): 				XYZ -> llh using the datum designed for using 
    							datum() as input
    ellecc(llh,vector,optional)
    eccell(llh,vector, swich ):			        converting a vector in xyz coord. system to local                                                        enu system in m
    xyz2enumatr(refarray,swich):			rotation matrix to convert xyz coord. system to local enu system or vice versa:w
    greatcircd(coor1,coor2):				distance along great circle
    M:
    platePoles(plate,units="ry")
    rotpole(plate,xyz,swich):
    appendSpherical_np(xyz)
"""

import pyproj as proj
from pyproj import CRS, Transformer

import numpy as np
import scipy as sp

EARTH_ERAD = 6378137.0

# itrf2008=proj.Proj("+proj=geocent +ellps=GRS80 +units=m +no_defs") # itrf2008
# lonlat=proj.Proj(init="EPSG:4326")
# isn2004=proj.Proj("+proj=lcc +lat_1=64.25 +lat_2=65.75 +lat_0=65 +lon_0=-19 +x_0=1700000 +y_0=300000 +no_defs +a=6378137 +rf=298.257222101 +to_meter=1") #isnet 2004
itrf2008 = CRS("EPSG:5332")
wgs84 = CRS("EPSG:4326")
lonlat = CRS("EPSG:4326")
isn2004 = CRS("EPSG:5322")
isn93 = CRS("EPSG:3057")

itrf08towgs84 = Transformer.from_crs(itrf2008, wgs84, always_xy=True )
itrf08toisn04 = Transformer.from_crs(itrf2008, isn2004)
itrf08toisn03 = Transformer.from_crs(itrf2008, isn93)


### core functions for manipulating and working with vectors on earth
def loc_to_geod(loc_coord):


    """

    Function to convert a location to geodetic coordinates. It calculates the following:
    colat           - Colatitude (used so that loc_coord and pos_geod - can be the same arrays)
    colat_trun      - Coltitude truncated to the nearest 10 asecs - Used to compute small circle radius
    loc_coord(3)    - longitude (rad), latitude (rad) and height (m)
    pos_geod(3)     - Distance from equator, distance from Greenwich - meridian, and height above ellipsiod (all m)


    Args:
        loc_coord: numpy array longitude (rad), latitude (rad) and height (m)

    Returns:
        pos_geod: numpy array distance from equator, distance from Greenwich - meridian, and height above ellipsiod (all m)

    """
    pos_geod = np.zeros(3)

    #  Distance from equator
    colat = np.pi / 2 - loc_coord[1]
    # ! 20000 = 1/10 in radians
    colat_trun = np.round(colat * 20000) / 20000

    pos_geod[0] = loc_coord[1] * EARTH_ERAD

    # Distance from Greenwich meridian along small circle
    if loc_coord[0] < 0:
        pos_geod[1] = (2 * np.pi + loc_coord[0]) * EARTH_ERAD * np.sin(colat_trun)
    else:
        pos_geod[1] = loc_coord[0] * EARTH_ERAD * np.sin(colat_trun)

    # Height above ellipssiod

    pos_geod[2] = loc_coord[2]

    return pos_geod


def geod_to_loc(pos_geod):
    """

    Function to convert geodetic coordinates to a location. It calculates the following:
    colat           - Colatitude (used so that loc_coord and pos_geod
                    - can be the same arrays)
    colat_trun      - Coltitude truncated to the nearest 10 asecs
                    - Used to compute small circle radius
    loc_coord(3)    - longitude (rad), latitude (rad) and
                    - height (m)
    pos_geod(3)     - Distance from equator, distance from Greenwich
                    - meridian, and height above ellipsiod (all m)

    Returns:
        loc_coord: numpy array longitude (rad), latitude (rad) and height (m)
    """

    loc_coord = np.zeros(3)

    #  Get the colat and loc_coord(1)
    loc_coord[1] = pos_geod[0] / EARTH_ERAD

    colat = np.pi / 2 - loc_coord[1]
    #   ! 20000 = 1/10" in radians
    colat_trun = np.round(colat * 20000) / 20000

    # Now get longitude
    loc_coord[0] = pos_geod[1] / (EARTH_ERAD * np.sin(colat_trun))
    if loc_coord[0] > np.pi:
        loc_coord[0] = loc_coord[0] - 2 * np.pi

    # Height above ellipssiod

    loc_coord[2] = pos_geod[2]

    return loc_coord


def gdatum(datum="IGS08"):
    """
    Function that extracts Datum information from bernese DATUM file. It returns the major axes ($ae), minor axes ($be), scale ($sc), shift (@dx), rotation (@rx)
    (Planned function - yet to be implemented)
    Args:
        datum

    Returns:
        To be Implemented
    """
    raise NotImplemented


def xyzell(xyzcoor, datum=itrf2008, radians=True):
    """
    Function to convert xyz coordinates to llh. It uses the pyproj library

    Args:
        x,y,z => long,lat,height
        usage: xyzell(xyzcoor,datum)
        datum: reference to datum definition
        xyzcoor: reference to a xyz coordinates
        returns reference to a list
         ( LONGITUDE, LATITUDE, HEIGHT(over ellipsoid))

    Returns:
        np.array(lon,lat,height)
    """

    return np.array(proj.transform(datum, lonlat, *xyzcoor, radians=radians))


def ellecc(llh, vector):
    """

    Function that converts a list of vectors in ECEF coordinates to the local enu coordinate system at llh.

    Args:
        enu displacement -> xyz displacement
        usage: &ellecc(llh,vector,optional)
        llh: reference to coordinates in ll or llh
        vector: reference to vector array in neu coor

    Returns:
        displacement in enu coordinate system.

    (Planned function, not yet implemented)

    """
    raise NotImplemented


def eccell(ll, vector):
    """
    Function that converts xyz displacement to enu displacement for a list of vectors in ECEF coordinates to the local enu coordinate system at llh.
    xyz displacement -> enu displacement

    Args:
        llh: list of coordinates in ll or llh line vector (long,lat,[height]): (rad,rad,m)
        vector: A list of vector in ECEF coordinates (m)

    Returns:
        List of the vectors in vectors in a local enu reference system.
    """

    if len(ll) != len(vector):
        raise ValueError

    rmat = xyz2enumatr(ll)

    if len(vector.shape) == 1:  # has to be numpy array
        vector = np.expand_dims(vector, axis=0)

    return np.squeeze(
        [np.transpose(x * y.reshape(3, 1)) for (x, y) in zip(rmat, vector)]
    )


def xyz2enumatr(ll):
    """
    Function that calculates a tranformation matrix to convert xyz to local enu coordinates system enu or viceversa.

    Args:
        ll list of lon,lat (radians) locations to compute the transformation matrix

    Returns:
        list of transformation matrixes for the corresponding ll values

    """

    f = lambda x: [
        np.sin(x),
        np.cos(x),
    ]  # For calculating sin(lambda),shi(phi),cos(lambda),cos(phi)

    # Creating the rotation matrix [ [-slmb, clmb, 0], [-sphi*clmb, -sphi*slmb, cphi], [cphi*clmb, cphi*slmb, sphi]]
    # a = (slmb, sphi, clmb, cphi)
    m = lambda x: [
        np.matrix(
            [
                [-a[0], a[2], 0],
                [-a[1] * a[2], -a[1] * a[0], a[3]],
                [a[3] * a[2], a[3] * a[0], a[1]],
            ]
        )
        for a in x
    ]

    if len(ll.shape) == 1:  # In case the array is only 1 demintion
        ll = np.expand_dims(ll, axis=0)

    # calculating
    [slmb, sphi], [clmb, cphi] = f(ll[:, 0:2].transpose())

    rmat = m(zip(slmb, sphi, clmb, cphi))
    return rmat


def greatcircd(coor1, coor2, ellps="GRS80"):
    """
    Function that calculates the  distance along great circle between two points on the Earth's surface.

    Args:
        coor1: reference to one point on earth's surface given in long, lat radians.
        coor2: reference to another point on earth's surface given in long, lat radians.

    Returns:
        great circle distance between coor1 and coor2

    """

    geod = proj.Geod(ellps=ellps)

    print(geod)


def M():
    """
    Function that calculates the Matrix of convenience in the local enu coordinate system.

    Args:

    Returns:

    """

    M = np.matrix("1, 0, 0, -1, 0, 0; 0, 1, 0, 0, -1, 0; 0, 0, 1, 0, 0, -1")
    return M


def platePoles(plate, units="ry"):
    """
    Function that calculates the angular velocities of tectonic plates as published in altamimi et al (2012).

    Args:
        plate: plate name (string)
        units: angular velocity units (string)

    Returns:
        pole: angular velocity of the plate (mas/Y)

    """
    mas2dy = 1.0 / (60 * 60) / 1000  # converting mas/Y to deg/Y

    # The plate name
    plates = {
        "EURA": 0,
        "NOAM": 1,
        "NAZC": 2,
        "INDI": 3,
        "NUBI": 4,
    }

    # ITRIF2008 Absolute Plate rotation poles From Table 3 in altamimi etal2012
    poles = np.array(
        [
            [-0.083, -0.534, 0.750],
            [0.035, -0.662, -0.100],
            [-0.330, -1.551, 1.625],
            [1.232, 0.303, 1.540],
            [0.095, -0.598, 0.723],
        ]
    )  # [mas/y]

    pole = poles[plates[plate], :]

    if units == "masy":
        return pole

    # converting to deg/Y
    pole = pole * mas2dy
    if units == "degy":
        return pole
    if units == "degMy":
        return pole * 1000000

    if units == "ry":
        return np.radians(pole)  # to radians/year


def rotpole(wxwywz, xyz, plate=None):
    """
    Function that calculates the nnr-nuvela velocity of a given plate in geocentric coordinates.

    If plate is passed:
        Calculate the nnr-nuvela velocity of a given plate in geocentric coordinates.
    If plate is not passed:
        Calculate the velocity  (cross product) in a given point in ECEF coordinates with for a given rotation pole wxwywz

    Args:
        plate: plate name (string) or None
        wxwywz: angular velocity of the plate (mas/Y)
        xyz: reference to a point in ECEF coordinates

    Returns:
        nnr-nuvela velocity of a given plate in geocentric coordinates

    """
    if plate != None:
        pass
    else:
        if xyz is not None and wxwywz is not None:
            return sp.cross(wxwywz, xyz)
        else:
            print("you need to pass a location in rad and angular velocity")


def appendSpherical_np(xyz):
    """
    Function that calculates the spherical coordinates of a point.

    Args:
        xyz: reference to a point in ECEF coordinates

    Returns:
        spherical coordinates
    """

    ptsnew = np.hstack((xyz, np.zeros(xyz.shape)))
    xy = xyz[:, 0] ** 2 + xyz[:, 1] ** 2
    ptsnew[:, 3] = np.sqrt(xy + xyz[:, 2] ** 2)
    ptsnew[:, 4] = np.arctan2(
        np.sqrt(xy), xyz[:, 2]
    )  # for elevation angle defined from Z-axis down
    # ptsnew[:,4] = np.arctan2(xyz[:,2], np.sqrt(xy)) # for elevation angle defined from XY-plane up
    ptsnew[:, 5] = np.arctan2(xyz[:, 1], xyz[:, 0])
    return ptsnew



# -*- coding: utf-8 -*-

# ------------------------------- #
#
# getSeptentrio.py 0.5
# Code made by bgo@vedur.is modified from fjalar@vedur.is
# Iceland Met Office
# 2018
#
# ------------------------------- #

#----------# IMPORT LIBRARIES #-----------#
# Common modules
import sys, signal, argparse, os, time
import os.path as path
from datetime import datetime as datetime
import urllib.request, urllib.parse, urllib.error, shutil, subprocess

import gtimes.timefunc as gt

import os
import shutil
import sys
import re
from ftplib import FTP
from datetime import datetime
import gtimes.timefunc as gpstime

def check_for_file(doyo, year, archive_path, station_id):
    """
    Checks if the file for a given day number of the year and year exists in the data archive.

    Args:
        doyo (int): Day number of the year.
        year (int): Year to check the file for.
        archive_path (str): Path to the archive where files are stored.
        station_id (str): Station ID to check the file for.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    file_name = get_file_name(doyo, year, station_id)
    full_path = os.path.join(archive_path, file_name)
    print(f" >> Checking for file {full_path}")
    file_found = os.path.isfile(full_path)
    if not file_found:
        print(" ---> File not found. Will be downloaded.")
    return file_found

def download_file(file_name, filedir_name, location_dir, ip_number, ip_port):
    """
    Downloads a file from a specified IP address and port using FTP into a given directory.

    Args:
        file_name (str): The name of the file to download.
        filedir_name (str): The directory name on the FTP server where the file is located.
        location_dir (str): The local directory where the file will be saved.
        ip_number (str): The IP address of the FTP server.
        ip_port (int): The port number of the FTP server.

    Returns:
        bool: True if the file was successfully downloaded, False otherwise.
    """
    file_downloaded = False
    os.chdir(location_dir)
    try:
        ftp = FTP()
        ftp.connect(ip_number, ip_port)
        ftp.login('anonymous')
        # Special handling for certain station names
        if re.search('ROTH|SVIN|SVIE', file_name):
            ftp.set_pasv(False)
        ftp.cwd(f'/DSK1/SSN/LOG1_15s_24hr/{filedir_name}')
        with open(file_name, 'wb') as fhandle:
            ftp.retrbinary(f'RETR {file_name}', fhandle.write)
        file_downloaded = True
    except Exception as e:
        print(f"Failed to download {file_name}: {e}")
    finally:
        ftp.quit()
    return file_downloaded

def archive_file(file_name, formatted_file_name, archive_path_dict, tmp_dir, year):
    """
    Moves the downloaded file from a temporary directory to the archive directory.

    Args:
        file_name (str): The name of the file in the temporary directory.
        formatted_file_name (str): The name for the file in the archive directory.
        archive_path_dict (dict): A dictionary containing the components of the archive path.
        tmp_dir (str): The temporary directory where the file is initially downloaded.
        year (int): The year associated with the file.

    Returns:
        bool: True if the file was successfully archived, False otherwise.
    """
    file_archived = False
    if not os.path.isdir(archive_path_dict['full_path']):
        make_directory(archive_path_dict)
    source_path = os.path.join(tmp_dir, file_name)
    destination_path = os.path.join(archive_path_dict['full_path'], formatted_file_name)
    if not os.path.isfile(destination_path):
        shutil.move(source_path, destination_path)
        file_archived = True
    else:
        print(f" >> File {destination_path} already exists!")
    return file_archived

def get_archive_path(doyo, year, station_id):
    """
    Generates the path to the archive based on the day of the year, year, and station ID.

    Args:
        doyo (int): Day of the year.
        year (int): Year for the archive path.
        station_id (str): Station ID for the archive path.

    Returns:
        dict: A dictionary containing the components and the full path to the archive.
    """
    month_dict = {
        '01': 'jan', '02': 'feb', '03': 'mar', '04': 'apr',
        '05': 'may', '06': 'jun', '07': 'jul', '08': 'aug',
        '09': 'sep', '10': 'oct', '11': 'nov', '12': 'dec'
    }
    file_date = gpstime.toDatetime(f"{year}-{doyo}", "%Y-%j")
    month = month_dict[file_date.strftime("%m")]
    full_path = f"/data/{year}/{month}/{station_id}/15s_24hr/raw/"
    archive_path_dict = {
        'root': '/data',
        'year': year,
        'month': month,
        'station_id': station_id,
        'frequency': '15s_24hr',
        'data_type': 'raw',
        'full_path': full_path
    }
    return archive_path_dict

def get_file_name(doyo, year, station_id):
    """
    Generates a file name based on the day of the year, year, and station ID.

    Args:
        doyo (int): Day of the year.
        year (int): Year for the file name.
        station_id (str): Station ID to include in the file name.

    Returns:
        str: The generated file name.
    """
    file_date = gpstime.toDatetime(f"{year}-{doyo}", "%Y-%j")
    return f"{station_id}{year}{file_date.strftime('%m%d')}0000a.sbf.gz"

def make_directory(archive_path_dict):
    """
    Creates the archive directory based on the provided path dictionary.

    Args:
        archive_path_dict (dict): A dictionary containing the components of the archive path.

    Returns:
        None: This function does not return any values but creates directories as needed.
    """
    root = archive_path_dict['root']
    for folder in [archive_path_dict['year'], archive_path_dict['month'], archive_path_dict['station_id'], archive_path_dict['frequency'], archive_path_dict['data_type']]:
        root = os.path.join(root, folder)
        if not os.path.isdir(root):
            os.makedirs(root)

def program_info_screen():
    """
    Prints software information.
    """
    print('')
    print("Copyright (c) 2016 Icelandic Met Office")
    print("getLeica 0.1 (Jul 2016)")
    print('')

def exit_gracefully(signum, frame):
    """
    Handles a graceful exit when Ctrl-C is pressed.

    Args:
        signum (int): Signal number.
        frame (frame): Current stack frame.

    Returns:
        None: This function does not return any values but exits the program if the user confirms.
    """
    original_sigint = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, original_sigint)
    try:
        if input("\nReally quit? (y/n)> ").lower().startswith('y'):
            sys.exit(1)
    except KeyboardInterrupt:
        print('Ok ok, quitting')
        sys.exit(1)
    signal.signal(signal.SIGINT, exit_gracefully)

#def sync_data(sta, start=None, end=None, session="15s_24hr", ffrequency="24hr", afrequency="15s", clean_tmp=True, sync=False, compression=".gz"):
#    """
#    This is the main loop in the program.. and could be moved to the main loop: REFACTORING
#    
#    1) Takes in the number of days to check backwards
#    2) Loops backwards, calls check_for_file and collects the files that are missing in a list
#    3) Loops through the list and downloads the files into a temp directory
#    """
#    
#    from datetime import datetime as dt
#    from datetime import timedelta as td
#
#    import cparser
#    import re
#    import os
#    
#    from gtimes.timefunc import currDatetime
#
#    print(("Start: {}, end: {}".format(start, end)))
#
#    # Get date today
#    station_id = sta 
#    
#    # handling default Time
#    hoursession = re.compile("1hr",re.IGNORECASE)
#    if ffrequency.lower == "1hr": # for downloading hourly data
#        if not end:
#            end =  (dt.now() - td(hours=1))
#        end = end.replace(minute=0, second=0, microsecond=0)
#    else: # Daily data
#        if not end:
#            end = currDatetime(-1)
#        end = end.date()
#
#    if hoursession.search(session): # for downloading hourly data
#        if not start:
#            start = end - td(hours=24)
#        start = start.replace(minute=0,second=0, microsecond=0)
#    else: # and for daily
#        if not start:
#            start = end - td(days=10)
#        start = start.date()
#
#
#
#    today = gt.currDatetime()
#    today_fancy = today.strftime("%A %d. %B %Y")
#    today_start_time = today.strftime("%H:%M:%S")
#
#    # Time the process
#    start_time = time.time()
#    
#
#    download_file_dict = {}
#    tmp_dir = "/home/gpsops/tmp/download/{}/".format(station_id)
#
#    predir="/DSK1/SSN/"
#    #session="LOG1_15s_24hr"
#    sessionn = "0"
#    sessionl = "a"
#    suffix = "sbf.gz"
#
#
#    parser = cparser.Parser()
#    station_info = parser.getStationInfo(station_id.upper())
#    if station_info:
#        ip_number = station_info["router"]["ip"]
#        print(ip_number)
#        ip_port = station_info['receiver']['ftpport']
#        print(("USING CPARSER: station {0} has IP:FTP_PORT {1}:{2}".format(station_id, ip_number, ip_port)))
#    else:
#        print('')
#        print(('__main__ ERROR: Unknown station ID {0}. Use "info" to query station info.'.format(station_id)))
#        exit()
#
#    # Quick fix as ftp may not be passive for imo APN server should go into config
#    regexp = re.compile("10\.4\.1")
#
#    if regexp.search(ip_number):
#        pasv = False
#        print(("SET pasv=False for IP: {}".format(ip_number)))
#    else:
#        pasv = True
#
#    # Boolean switches for function execution success
#    downloaded_files_dict = {}
#    file_archived = False
#
#    
#    #--------------------------------#
#    # 2) Create the temp directory if not already existing
#    #--------------------------------#
#
#    if os.path.isdir(tmp_dir):
#        pass
#    else:
#        print("STATUS > Temp directory {0} for downloading is missing".format(tmp_dir))
#        print("         creating it ....")
#        os.mkdir(tmp_dir)
#
#
#    #--------------------------------#
#    # 3) Print report header
#    #--------------------------------#
#
#    print("Program run on {}".format(today_fancy))
#    print("Time started: {}".format(today_start_time))
#    print("Current day number (doy): {}".format(gt.DayofYear())) 
#    print("checking for {0} sessions: from {1} to {2}".format(session, start, end))
#
#
#    #print(make_file_name(station_id, gt.currDatetime(-2), session, receiver_type="POLARX5", ftype="IMOstd"))
#    stringformat = "#datelist"
#    lfrequency = "1D"
#    file_datetime_list = gt.datepathlist(stringformat, lfrequency, starttime=start, endtime=end, datelist=[], closed=None)
#
#    stringformat = "/data/%Y/#b/{0}/{1}/raw/{0}%Y%m%d%H00a.sbf{2}".format(sta, session, compression)
#    archive_file_list = gt.datepathlist(stringformat, lfrequency, datelist=file_datetime_list, closed=None)
#    print(archive_file_list)
#
#    stringformat1 = "{}#Rin2_{}".format(sta,compression)
#    IGS_file_name_list = gt.datepathlist(stringformat1, lfrequency, datelist=file_datetime_list, closed=None)
#    file_date_dict = dict(list(zip(file_datetime_list, list(zip(archive_file_list, IGS_file_name_list)) )))
#
#    # checking if file is in archive
#    print("constructing a list for files missing from archive")
#    #missing_file_dict = { key:value for (key,value) in file_date_dict.items() if not os.path.isfile(value[0]) }
#    missing_file_dict = { key:value for (key,value) in list(file_date_dict.items()) }
#    
#    if len(list(missing_file_dict.keys())) == 0:
#        print('*-------------------------------------------------------------------------*')
#        print("")
#        print("STATUS > Archive is up to date.")
#        sync = False
#    else:
#        print('*-------------------------------------------------------------------------*')
#        print("")
#        print("STATUS > Missing files: ")
#        for key,value in list(missing_file_dict.items()):
#            print(("{0}: {1}".format(key,value) ))
#        print("")
#
#
#    if sync:
#
#
#        print((Session(session)[1]))
#        stringformat2 = "{}{}/%y%j/".format(predir,Session(session)[1])
#        print(stringformat2)
#        remote_path_list = gt.datepathlist(stringformat2, lfrequency, datelist=list(missing_file_dict.keys()), closed=None) 
#
#        # packing files and paths in a dict for downloading
#        download_file_dict = dict(list(zip( list(zip(*list(missing_file_dict.values())))[1], remote_path_list )))
#
#        # Connecting to server
#        ftp = ftp_open_connection(ip_number,ip_port,pasv=pasv, timeout=10)
#        print( " >> Now downloading and archiving missing files ...")
#        downloaded_files_list = ftp_download(download_file_dict, tmp_dir, clean_tmp=clean_tmp, ftp=ftp, ftp_close=True)
#
#        downloaded_files_dict = dict(list(zip( missing_file_dict, downloaded_files_list)))
#
#        #archiving
#        if downloaded_files_dict:
#            for ddate , tmp_file in list(downloaded_files_dict.items()):
#                print(("File to archive {}".format(os.path.basename(tmp_file))))
#                tmp_file_size = os.path.getsize(tmp_file)
#
#                if os.path.isfile(missing_file_dict[ddate][0]):
#                    archive_file_size = os.path.getsize(missing_file_dict[ddate][0])
#
#                    if tmp_file_size == archive_file_size:
#                        print(("Files dated {0}:\n {1} and {2}\nhave the same size {3} bytes.\nAborting"
#                                .format(ddate, tmp_file, missing_file_dict[ddate][0], tmp_file_size)))
#                else:
#                    archive_path, archive_file_name =  os.path.split(missing_file_dict[ddate][0]) 
#                    if not os.path.isdir(archive_path):
#                        print(("Directory {0} does not exist creating it ...".format(archive_path)))
#                    print(("Move file dated {0} from {1} to {2}".format(ddate, tmp_file, missing_file_dict[ddate][0])))
#                    os.rename(tmp_file,missing_file_dict[ddate][0])
#                    archive_file_size = os.path.getsize(missing_file_dict[ddate][0])
#                    if tmp_file_size == archive_file_size:
#                        print("File succsessfully moved")
#
#            
#        if file_archived:
#            print("STATUS > File downloaded and archived!")
#        else:
#            print("STATUS > Error in downloading and archiving. File might be missing on the receiver.")
#
#    #--------------------------------#
#    # 5) Output end time
#    #--------------------------------#
#
#    today_end_time = today.strftime("%H:%M:%S")
#    print("")
#    print("PROCESS DURATION: {0:.2f} seconds".format(time.time()-start_time))
#    return downloaded_files_dict
#
#def Session(session):
#    """
#    """
#
#    Session = {
#            "15s_24hr": ("a","LOG1_15s_24hr"),
#            "1Hz_1hr":  ("b","LOG2_1Hz_1hr"),
#            }
#
#    return Session[session]
#
#def make_file_name(station_id, day, session="15s_24hr", receiver_type="POLARX5", ftype="IMOstd", compression=".gz"):
#    """
#    """
#
#    import gtimes.timefunc as gt
#    import re
#
#    file_name = ""
#    suff_dict={
#            "POLARX5": "sbf",
#            }
#
#    
#    daysession = re.compile("24hr",re.IGNORECASE)
#    hoursession = re.compile("1hr",re.IGNORECASE)
#    
#    if ftype == "IMOstd":
#
#        print(("Session name: {}".format(session)))
#        if daysession.search(session): filedate = day.strftime("%Y%m%d0000a")
#
#        if hoursession.search(session): filedate = day.strftime("%Y%m%d%H00b")
#
#        file_name = "{0}{1}.{2}{3}".format(station_id, filedate, suff_dict[receiver_type], compresssion)
#
#    if not file_name:
#        print(("The fromat {0} is unknown".format(ftype)))
#        
#    return file_name
#
#
#def ftp_open_connection(ip_number,ip_port,pasv=True, timeout=10):
#
#    """
#        open ftp connection
#    """
#
#    from ftplib import FTP
#    
#    ## Try to connect to the server
#
#   # TEMP stuff sometimes we need passive ftp will go to config
#
#    try:
#        print("Connection to station...")
#        ftp = FTP()
#        ftp.connect(ip_number, ip_port, timeout=timeout)
#        ftp.login('anonymous')
#        ftp.set_pasv(pasv)
#        print("Connection successful!")
#    except: 
#        print("Connection failed")
#
#        ftp = None 
#
#    return ftp
#
#
#def ftp_download(files_dir_to_download_dict, local_dir, clean_tmp=True,
#                 ftp=None, ip_number=None, ip_port=None,  pasv=True, ftp_close=True):
#    """
#    download a list of files from an ftp server
#    """
#
#    import os
#    import re
#    from ftplib import FTP
#
#    if not ftp:
#        ftp = ftp_open_connection(ip_number ,ip_port, pasv=pasv)
#
#    if not ftp:
#        print(( "Can't connect to {}:{}, nothing downloaded".format(ip_number ,ip_port) ))
#        return []
#  
#    downloaded_files = []
#    remote_file_size = {}
#
#    # Execute file download if connection was succsesfull 
#    for file_name, remote_dir in sorted(list(files_dir_to_download_dict.items()),reverse=True):
# 
#        print("=====================================")
#        print(("File name: {}".format(file_name)))
#        print(("Remote directory: {}".format(remote_dir)))
#        print(("Download directory: {}".format(local_dir)))
#        print("-------------------------------------")
#        print(">")
#
#
#        local_file = "{0}{1}".format(local_dir,file_name)
#        if clean_tmp is True and os.path.isfile(local_file):
#            os.remove(local_file)
#
#        remote_file= "{0}{1}".format(remote_dir,file_name)
#    
#        offset = 0 
#        if os.path.isfile(local_file):
#            # check how much has alread been downloaded
#                offset = os.path.getsize(local_file)
#
#        
#        # Download the file
#        print('Downloading ' + file_name)
#
#        # remote_file
#        try:
#            remote_file_size = ftp.size(remote_file)
#        except:
#                remote_file_dict = ftp_list_dir([remote_dir], ftp=ftp, ftp_close=False)
#                print(remote_file_dict)
#                basename = file_name[0:11]
#                base_regexp = re.compile(basename)
#
#                
#                print(("File {} not on recever listing remote files".format(remote_file)))
#                print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
#                for rdir, rfile_list in list(remote_file_dict.items()):
#                    for rfile in rfile_list:
#                        print(rfile) 
#                print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
#                print(">")
#
#                for rdir, rfile_list in list(remote_file_dict.items()):
#                    for rfile in rfile_list:
#                        file_name=rfile.split()[-1]
#                        
#                        local_file = "{0}{1}".format(local_dir,file_name)
#                        if os.path.isfile(local_file):
#                            # check how much has alread been downloaded
#                            offset = os.path.getsize(local_file)
#    
#                        if base_regexp.search(file_name):
#                            remote_file="{0}{1}".format(rdir,file_name)
#                            print(("Found file {0} on receiver will download to \n {1}".format(remote_file, local_file) +
#                                  " File will not be archived automatically." )) 
#                            remote_file_size = ftp.size(remote_file)
#                            diff = _download_with_progressbar(ftp, remote_file,local_file, remote_file_size, offset=offset)
#                        else:
#                            print(("Did not find any file matching {0} on the receiver, \n".format(basename) + 
#                                  "check if the receiver or station info is configured correctly"))
#                    
#                
#                continue
#
#        diff = _download_with_progressbar(ftp, remote_file,local_file, remote_file_size, offset=offset)
#        print(("Difference between remote and downloaded file: {0:d}".format(diff)))
#        if diff == 0:
#            downloaded_files.append(local_file)
#
#    if ftp_close:
#        ftp.close()
#
#    return downloaded_files 
#
#
#def ftp_list_dir(dir_list, ftp, ip_number=None, ip_port=None, pasv=True, ftp_close=True):
#    """
#    """
#
#    import os
#    import re
#    from ftplib import FTP
#
#    if not ftp:
#        ftp = ftp_open_connection(ip_number ,ip_port, pasv=pasv)
#
#    if not ftp:
#        print(( "Can't connect to {}:{}, nothing downloaded".format(ip_number ,ip_port) ))
#        return []
#    
#    remote_file_dict = {}
#    for remote_dir in dir_list:
#        remote_dir_list=[]
#        ftp.dir(remote_dir, remote_dir_list.append)
#        remote_file_dict[remote_dir] = remote_dir_list
#    
#
#    if ftp_close:
#        ftp.close()
#
#    return remote_file_dict
#
#
#def is_gz_file(filepath):
#    """
#    Check if a file is a gzip file
#    """
#
#    import binascii
#    
#    with open(filepath, 'rb') as test_f:
#        return binascii.hexlify(test_f.read(2)) == b'1f8b'
#
#
#def _download_with_progressbar(ftp, remote_file,local_file, remote_file_size, offset=0):
#    """
#    Download a file using a process bar. 
#    Returns a the difference in bytes between the remote file and the downloaded file
#    """
#
#    import progressbar
#    
#    progress = progressbar.AnimatedProgressBar(start=offset,end=remote_file_size, width=50)
#
#    with open(local_file, 'ab') as f:
#        def callback(chunk):
#            f.write(chunk)
#            progress + len(chunk)
#
#            # Visual feedback of the progress!
#            progress.show_progress()
#
#        ftp.retrbinary('RETR {0}'.format(remote_file), callback,rest=offset)
#        print("")
#    # print( remote_file_size )
#    local_file_size = os.path.getsize(local_file)    
#
#    return local_file_size-remote_file_size
#
#
#def program_info_screen():
#    ''' Print software info.'''
#    # Only splash screen info here
#
#    current_func = sys._getframe().f_code.co_name + '() >> '
#
#    print('')
#    print("Copyright (c) 2016 Icelandic Met Office")
#    print("getLeica 0.1 (Jul 2016)")
#    print('')
#
#def exit_gracefully(signum, frame):
#    ''' Exit gracefully on Ctrl-C '''
#
#    current_func = sys._getframe().f_code.co_name + '() >> '
#
#    # restore the original signal handler as otherwise evil things will happen
#    # in raw_input when CTRL+C is pressed, and our signal handler is not re-entrant
#    signal.signal(signal.SIGINT, original_sigint)
#
#    try:
#        if input("\nReally quit? (y/n)> ").lower().startswith('y'):
#            sys.exit(1)
#
#    except KeyboardInterrupt:
#        print('Ok ok, quitting')
#        sys.exit(1)
#
#    # restore the exit gracefully handler here
#    signal.signal(signal.SIGINT, exit_gracefully)
#
#    # Method borrowed from:
#    # http://stackoverflow.com/questions/18114560/python-catch-ctrl-c-command-prompt-really-want-to-quit-y-n-resume-executi
#
##def main():
##    ''' main '''
##
##
##    import re
##    from datetime import datetime as dt
##    from datetime import timedelta as td
##
##    from gtimes.timefunc import currDatetime
##
##    #start = end = None
##    dstr="%Y%m%d-%H%M" # Default input string
##
##    # Display some nice program info
##    program_info_screen()
##
##    # Instantiate argparser
##    parser = argparse.ArgumentParser()
##
##    # Setup the argument parser
##    parser.add_argument('Stations', nargs='+',
##                        help='List of stations to download')
##    parser.add_argument('-D', '--days', type=int, default=10,
##                        help="Number of days back to check for data.")
##    parser.add_argument('-s','--start', type=str , default=None,
##                        help="Start date, format ""%%Y%%m%%d-%%H%%M"".")
##    parser.add_argument('-e','--end', type=str , default=None, 
##                        help="End date, format ""%%Y%%m%%d-%%H%%M"".")
##    parser.add_argument('-se', '--session', type=str, default='15s_24hr',
##                        help="Data sampling sessions. Default is 15s_24hr, 1Hz_1hr, 20HZ_1hr.")
##    parser.add_argument('-comp', '--compression', type=str, default='.gz',
##                        help="Sefine compression type")
##    parser.add_argument('-ffr', '--ffrequency', type=str, default='',
##                        help="Data file frequency, defaults to empty string and will try to determain it from session string")
##    parser.add_argument('-afr', '--afrequency', type=str, default='',
##                        help="Aqusition frequency, defaults to empty string and will try to determain it from session string")
##    parser.add_argument('-sy', '--sync', action='store_true',
##                        help='Sync new or partal files from source.')
##    parser.add_argument('-cl', '--clean_tmp', action='store_true',
##                        help='Clean download directory and start over on partly finished downlands')
## 
##    # Fetch the arguments
##    args = parser.parse_args()
##
##    # defining sub-periods
##    if args.start: # start the plot at
##       args.start = dt.strptime(args.start,dstr)
##    if args.end: # end the plot at
##        args.end = dt.strptime(args.end,dstr)
##
##    if not args.start and args.days:
##       args.start = currDatetime(days=-args.days, refday=dt.now())
##
##    # handling aqcustions frequency 
##    if args.afrequency:
##        pass
##    else:
##        args.afrequency = args.session.split("_")[0]
##    
##    # Handling file frequency
##    if args.ffrequency:
##        pass
##    else:
##        args.ffrequency = args.session.split("_")[1]
##
##
##
##    stations = args.Stations
##    kwargs = vars(args) 
##
##    del(kwargs['days'])
##    del(kwargs['Stations'])
##    
##    #print("ARGS: {}".format(kwargs))
##    for sta in stations:
##        sync_data(sta, **kwargs)
##
##
##if __name__ == '__main__':
##    # This is used to catch Ctrl-C exits
##    original_sigint = signal.getsignal(signal.SIGINT)
##    signal.signal(signal.SIGINT, exit_gracefully)
##
##    main()



#!/usr/bin/python
"""
This module contains some useful functions for gps data processing such as:

convllh(llh,radians=True)
extractfromGamitBakf(cfile, stations)
savedisp(dataDict,fname=None, header="")
"""
import subprocess

def run_cmd(check_cmd):
    """
    Function that executes a system command and capture its output and return code. It runs a given system command using the subprocess module, waits for it to complete, and then captures its outputs and return code.

    Args:
        check_cmd (str): The system command to be executed.

    Returns:
        (int, str): A tuple containing two elements: 1. proc_check_returncode (int): The return code of the command. A return code of 0 typically indicates that the command was executed successfully. 2. proc_check_comm (bytes): The output of the command as a byte string. If the command produces no output, this will be an empty byte string.

    """
    ## Run command

    process = subprocess.Popen(check_cmd,shell=True,stdout=subprocess.PIPE)
    process.wait()

    proc_check_returncode = process.returncode
    proc_check_comm = process.communicate()[0].strip('\n'.encode())
    
    
    return proc_check_returncode,proc_check_comm

def run_syscmd(check_cmd,p_args):
    """
    Function executing a system command and capturing its output and return code.

    This function runs a given system command using the subprocess module,
    waits for it to complete, and then captures its output and return code.
    It also logs debug information if specified in the parameters.

    Parameters:
        check_cmd (str): The system command to be executed.
        p_args (dict): A dictionary containing various parameters, including 'debug' (bool): If True, print debug information to the console.

    Returns:
        (int, str):  A tuple containing two elements:  1. proc_check_returncode (int): The return code of the command. A return code of 0 typically indicates that the command was executed successfully, 255 indicates a timeout, and other values indicate failure. 2. proc_check_comm (str): The output of the command as a string. If the command  produces no output, this will be an empty string.
    
    """

    ## Run command
    import sys, re, subprocess, getopt, logging, copy, socket, types
    import gpstime

    currfunc=__name__+'.'+sys._getframe().f_code.co_name+' >>' # module.object name
    if p_args['debug']: print('%s Starting ...' % currfunc)

    process = subprocess.Popen(check_cmd,shell=True,stdout=subprocess.PIPE)
    process.wait()

    proc_check_returncode = process.returncode
    proc_check_comm = process.communicate()[0].strip('\n')
    
    
    #(3)# Make desicions according to output...
    if p_args['debug']: print('%s process.returncode:' % currfunc, proc_check_returncode)
    if p_args['debug']: print(currfunc,'process.communicate():\n---------------\n', proc_check_comm,'\n-------------')
    if proc_check_returncode == 0:
        if p_args['debug']: print("%s Command went well.." % currfunc)
    elif proc_check_returncode == 255:
        if p_args['debug']: print("%s Timeout..." % currfunc)
    else:
        if p_args['debug']: print("%s Command failed... " % currfunc)


def run_netcmd(netcmd, port, station, p_args):
    """
    Execute a network command based on the connection type and capture its output and return code.

    This function formats a network command string based on the connection type specified in the
    station dictionary and then executes the command using the `run_syscmd` function. It logs
    debug information if specified in the arguments.

    Args:
        netcmd (str): The network command template to be executed, with placeholders for the IP and port.
        port (int): The port number to be used in the network command.
        station (dict): A dictionary containing station configuration, including:
            - 'conn_type' (str): The connection type, which can be 'tunnel', 'sil', 'direct', or 'serial'.
            - 'rout_ip' (str): The routing IP address used for 'tunnel' connections.
            - 'recv_httpport' (str): The receiver HTTP port used for 'tunnel' connections.
            - 'sil_httpport' (str): The SIL HTTP port used for 'sil' connections.
            - 'recv_ip' (str): The receiver IP address used for 'direct' connections.
            - 'sil_computer' (bool): Indicates if a SIL computer is used for 'serial' connections.
        p_args (dict): A dictionary containing various arguments, including:
            - 'debug' (bool): If True, print debug information to the console.

    Returns:
        (int, str): A tuple where the first element is the return code of the command,
        and the second element is the output of the command as a string. A return code
        of 0 typically indicates that the command was executed successfully, and other
        values indicate failure. If the command produces no output, the second element
        will be an empty string.

    """
    currfunc=__name__+'.'+sys._getframe().f_code.co_name+' >>'
    if p_args['debug']: print('%s Starting ...' % currfunc)

    conntype=station['conn_type'].split(',') # methods defined by conn_type

    #(1)# define how to handle differennt communication types
    if 'tunnel' in conntype:
        # direct ip connection through tunnels 
        if p_args['debug']: print('%s connecting through %s on port %s' % (currfunc,station['rout_ip'], station['recv_httpport']))
        netcmd = netcmd % '%s:%s' % (station['rout_ip'], port )
        if p_args['debug']: print('%s netcmd=' % currfunc,netcmd)

    elif 'sil' in conntype:
        # direct ip connection through sil computer using port forwarding
        if p_args['debug']: print('%s NetRS/NetR9 receiver on a SIL station' % currfunc)
        netcmd = netcmd % '%s:%s' % ('localhost',station['sil_httpport'])
        if p_args['debug']: print('%s netcmd=' % currfunc,netcmd)

    elif 'direct' in conntype:
        # direct ip connection 
        if p_args['debug']: print('%s Direct connection' % currfunc)
        netcmd = netcmd % station['recv_ip']
        if p_args['debug']: print('%s netcmd=' % currfunc,netcmd)
				
    elif 'serial' in conntype:
        if station['sil_computer']:
            pass
					
        elif '3G' in station['conn_type']:
            pass

    #(2)# run the command and return the output to proc_checktemp_comm
    proc_checktemp_returncode, proc_checktemp_comm=run_syscmd(netcmd, p_args)
    
    # Return
    return proc_checktemp_returncode, proc_checktemp_comm


def checkError(proc_check_returncode, proc_check_comm, searchres, p_args):
    """
    Check for errors in the output of a system command.

    This function evaluates the return code and output of a system command to determine if an error occurred.
    It also searches for a specific string in the output to confirm successful execution.

    Args:
        proc_check_returncode (int): The return code from the system command.
        proc_check_comm (str): The output from the system command.
        searchres (bool): The result of a search operation, typically a boolean indicating if a specific string was found.
        p_args (dict): A dictionary containing various arguments, including:
            - 'debug' (bool): If True, print debug information to the console.

    Returns:
        int: An error code representing the status of the check:
            - 0 if no errors were found and the search result is True.
            - 11 if 'ERROR' is found in the output.
            - 1 for any other case indicating something is wrong.
    """
    currfunc = __name__ + '.' + sys._getframe().f_code.co_name + ' >>'
    if p_args.get('debug'): print('%s Starting ...' % currfunc)

    if proc_check_returncode == 0 and searchres:
        if p_args.get('debug'): print("%s all is OK" % currfunc)
        return 0
    elif 'ERROR' in proc_check_comm:
        if p_args.get('debug'): print("%s proc_check_comm returned: %s" % (currfunc, proc_check_comm))
        return 11
    else:
        if p_args.get('debug'): print("%s Something is wrong" % currfunc)
        print("ERROR in matching proc_check_comm =\n", proc_check_comm,
              "\nand proc_check_returncode=", proc_check_returncode,
              "\nsearchres=", searchres)
        return 1


def changeDictKeys(session, chdict, p_args):
    """
    Change the keys of dictionaries within a list according to a mapping.

    This function iterates over a list of dictionaries and renames keys according to a provided mapping.
    The changes are made in place.

    Args:
        session (list of dict): The list of dictionaries to process.
        chdict (dict): A dictionary containing the mapping of old keys to new keys.
        p_args (dict): A dictionary containing various arguments, including:
            - 'debug' (bool): If True, print debug information to the console.

    Returns:
        list of dict: The list of dictionaries with updated keys.
    """
    currfunc = __name__ + '.' + sys._getframe().f_code.co_name + ' >>'
    if p_args.get('debug'): print('%s Starting inplace key change ...\n--\nChanging keys: %s\n--\nIn: %s\n--\nTo: %s'
                                  % (currfunc, chdict.keys(), session, chdict.values()))

    for sess in session:
        for key in set(chdict.keys()) & set(sess.keys()):
            sess[chdict[key]] = sess.pop(key)

    return session

import socket
import re
import sys

def checkPort(host, port, p_args):
    """
    Check if a network port on a given host is open and accepting connections.

    Args:
        host (str): The hostname or IP address to check.
        port (int): The port number to check.
        p_args (dict): A dictionary containing various arguments, including:
            - 'debug' (bool): If True, print debug information to the console.

    Returns:
        int: Returns 0 if the connection to the port is successful, 1 otherwise.
    """
    # variable port needs to be of type int
    currfunc=__name__+'.'+sys._getframe().f_code.co_name+' >>'
    if p_args['debug']: print('%s Starting...' % currfunc)

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
        s.shutdown(2)
        #if p_args['debug']:
        print("%s Success connecting to %s on port: %s" % (currfunc, host, port))
        return 0
    except:
        #if p_args['debug']:
        print("%s Cannot connect to %s on port: %s" % (currfunc, host, port))
        return 1

import re

def getportdict(pref, station, p_args):
    """
    Extract port definitions from a station configuration dictionary based on a prefix.

    This function searches through the station configuration dictionary for keys that match
    a given prefix followed by '_port' and constructs a new dictionary with the results.

    Args:
        pref (str): The prefix used to identify port definitions in the station configuration.
        station (dict): A dictionary containing station configuration.
        p_args (dict): A dictionary containing various arguments, including:
            - 'debug' (bool): If True, print debug information to the console.

    Returns:
        (dict): A dictionary containing the extracted port definitions, where each key is the port type (derived from the original key) and the value is the port number as a string.
    """
    # Search string to check for defined ports
    searchstr = re.compile(r'\s*%s_(\w+)port=(\S+)\s*' % re.escape(pref))
    stationstr = ' '.join(['%s=%s' % (k, v) for k, v in station.items()])

    # Extracting the port definitions from the pattern pref_+port
    searchres = searchstr.findall(stationstr)
    recvport = dict(searchres)

    return recvport

import sys

def str_to_class(field):
    """
    Convert a string to a class object in the current module.

    This function takes the name of a class as a string and returns the class object with that name
    from the current module. If the class does not exist, it raises an AttributeError.

    Args:
        field (str): The name of the class to retrieve.

    Returns:
        class: The class object corresponding to the given name.

    Raises:
        AttributeError: If the class with the given name does not exist in the current module.
    """
    return getattr(sys.modules[__name__], field)




"""
A Python implementation of GPS related time conversions.

Copyright 2002 by Bud P. Bruegger, Sistema, Italy
mailto:bud@sistema.it
http://www.sistema.it

Modifications for GPS seconds by Duncan Brown

PyUTCFromGpsSeconds added by Ben Johnson

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place, Suite 330, Boston, MA  02111-1307  USA

GPS Time Utility functions

This file contains a Python implementation of GPS related time conversions.

The two main functions convert between UTC and GPS time (GPS-week, time of
week in seconds, GPS-day, time of day in seconds).  The other functions are
convenience wrappers around these base functions.

A good reference for GPS time issues is:
http://www.oc.nps.navy.mil/~jclynch/timsys.html

Note that python time types are represented in seconds since (a platform
dependent Python) Epoch.  This makes implementation quite straight forward
as compared to some algorigthms found in the literature and on the web.
"""

__author__ = "Benedikt G. Ofeigsson <bgo@vedur.is> edited from Duncan Brown <duncan@gravity.phys.uwm.edu>"
__date__ = "$Date: 2012/03/6"
__version__ = "$Revision: 1.6 $"[11:-2]
# $Source: /usr/local/cvs/lscsoft/glue/glue/gpstime.py,v $

import time
import math
import datetime

from sqlalchemy import Boolean

secsInWeek = 604800
secsInDay = 86400
gpsEpoch = (1980, 1, 6, 0, 0, 0)  # (year, month, day, hh, mm, ss)
epochTuple = gpsEpoch + (-1, -1, 0)


def dayOfWeek(year: int, month: int, day:int) -> int:

    """
    This function returns the day of the week for a given date

    Examples:
        >>> dayOfWeek(2023, 1, 3)
        2
        >>> dayOfWeek(2023, 12, 4)
        1

    Args:
        year: Four digit year "yyyy". Example 2013
        month: Month in integer from 1-12
        day: Day of month as integer 1-(28-31) depending on month

    Returns:
        int: A number representing the day of the week (0=Sunday, 1=Monday, .., 6=Saturday)
    
    """
    hr = 12  # make sure you fall into right day, middle is save
    t = time.mktime((year, month, day, hr, 0, 0, 0, 0, -1))
    pyDow = time.localtime(t)[6]
    gpsDow = (pyDow + 1) % 7
    return gpsDow


def gpsWeek(year:int, month:int, day:int) -> int:
    """
    Function that calculates (full) gpsWeek for given date (in UTC)

    Args:
        year: int
        month:
        day:

    Returns:
        gpsWeek
    """
    hr = 12  # make sure you fall into right day, middle is save
    return gpsFromUTC(year, month, day, hr, 0, 0)[0]


def julianDay(year:int, month:int, day:int) -> int:
    """ Function that calculates julian day=day since Jan 1 of year

    Args:
        year
        month
        day

    Returns:
        julian day

    """



    hr = 12  # make sure you fall into right day, middle is save
    t = time.mktime((year, month, day, hr, 0, 0.0, 0, 0, -1))
    julDay = time.localtime(t)[7]

    return julDay


def mkUTC(year:int, month:int, day:int, hour:int, min:int, sec:int) -> int:
    """
    Similar to python's mktime but for utc. Converts time list to time in utc

    Args:
        year:
        month:
        day:
        hour:
        min:
        sec:

    Returns:
        time tuple in UTC

    """
    spec = [year, month, day, hour, min, sec] + [0, 0, 0]
    utc = time.mktime(spec) - time.timezone
    return utc


def ymdhmsFromPyUTC(pyUTC) -> tuple:
    """
    Function that computes a tuple from a python time value in UTC

    Args:
        pyUTC:

    Returns:
        tuple from a python time
    """

    return time.gmtime(pyUTC)[:-3]


# def wtFromUTCpy(pyUTC, leapSecs=18):
#    """
#    Convenience function:
#    BGO 07.01.2013: changed the default leapsecond to 16
#    BGO 07.01.2019: changed the default leapsecond to 18
#         allows to use python UTC times and
#         returns only week and tow"""
#
#    ymdhms = ymdhmsFromPyUTC(pyUTC)
#    wSowDSoD = gpsFromUTC( *ymdhms)
#    return wSowDSoD[0:2]
#    #return ymdhms


def gpsFromUTC(year:int, month:int, day:int, hour:int, min:int, sec:int, leapSecs=None, gpst=True):
    """
    Converts UTC to: gpsWeek, secsOfWeek, gpsDay, secsOfDay

    a good reference is:  http://www.oc.nps.navy.mil/~jclynch/timsys.html

    This is based on the following facts (see reference above):

    GPS time is basically measured in (atomic) seconds since
    January 6, 1980, 00:00:00.0  (the GPS Epoch)

    The GPS week starts on Saturday midnight (Sunday morning), and runs
    for 604800 seconds.

    Currently, GPS time is 13 seconds ahead of UTC (see above reference).
    While GPS SVs transmit this difference and the date when another leap
    second takes effect, the use of leap seconds cannot be predicted.  This
    routine is precise until the next leap second is introduced and has to be
    updated after that.
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    SOW = Seconds of Week
    SOD = Seconds of Day

    Note:  Python represents time in integer seconds, fractions are lost!!!
    """
    import datetime as dt

    # Automatic or manually applyed leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(
            dTime=dt.datetime(year, month, day, hour, min, sec), gpst=gpst
        )

    secFract = sec % 1
    t0 = time.mktime(epochTuple)
    t = time.mktime((year, month, day, hour, min, sec, -1, -1, 0))
    # Note: time.mktime strictly works in localtime and to yield UTC, it should be
    #       corrected with time.timezone
    #       However, since we use the difference, this correction is unnecessary.
    # Warning:  trouble if daylight savings flag is set to -1 or 1 !!!
    t = t + leapSecs
    tdiff = t - t0
    gpsSOW = (tdiff % secsInWeek) + secFract
    gpsWeek = int(math.floor(tdiff / secsInWeek))
    gpsDay = int(math.floor(gpsSOW / secsInDay))
    gpsSOD = gpsSOW % secsInDay

    return (gpsWeek, gpsSOW, gpsDay, gpsSOD)


def UTCFromGps(gpsWeek:int, SOW:int, leapSecs:int=None, dtimeObj:Boolean=False):
    """
    Function that converts gps week and seconds to UTC.
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    see comments of inverse function!

    SOW = seconds of week
    gpsWeek is the full number (not modulo 1024)

    Args:
        gpsWeek:
        SOW:
        leapSecs:
        dtimeObj:

    Returns:
        datetime object in UTC time zone
    """
    import datetime as dt

    # Automatic or manually applied leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(dTime=(gpsWeek, SOW), gpst=True)

    secFract = SOW % 1
    epochTuple = gpsEpoch + (-1, -1, 0)
    t0 = time.mktime(epochTuple) - time.timezone  # mktime is localtime, correct for UTC
    tdiff = (gpsWeek * secsInWeek) + SOW - leapSecs
    t = t0 + tdiff
    if dtimeObj is True:
        # only deals with whole seconds
        return dt.datetime.fromtimestamp(t)
    else:
        (
            year,
            month,
            day,
            hh,
            mm,
            ss,
            dayOfWeek,
            julianDay,
            daylightsaving,
        ) = time.gmtime(t + secFract)
        # use gmtime since localtime does not allow to switch off daylighsavings correction!!!
        return (year, month, day, hh, mm, ss + secFract)


def GpsSecondsFromPyUTC(pyUTC, leapSecs=None):
    """
    Function that converts the python epoch to gps seconds
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    pyEpoch = the python epoch from time.time()

    Args:
        pyUTC:
        leapSecs:

    Returns:
        Integer representing the number of leapSeconds in a python epoch

    """
    import datetime as dt

    # Automatic or manually applied leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(dTime=dt.datetime(*ymdhmsFromPyUTC(pyUTC)), gpst=True)

    t = gpsFromUTC(*ymdhmsFromPyUTC(pyUTC - leapSecs))

    return int(t[0] * 60 * 60 * 24 * 7 + t[1])


def PyUTCFromGpsSeconds(gpsseconds):
    """
    Converts gps seconds to the python epoch. That is, the time that would be returned from time.time() at gpsseconds.

    Args:
        gpsseconds:

    Returns:
        Python epoch equivalent from the gps seconds.

    """

    pyUTC


def getleapSecs(dTime=None, gpst=True):
    """
    Function that retreives the leap seconds for a given datetime.

    Args:
        dTime: A Time string can be a datetime object, string or integer (either seconds since Epoch or GPSs)
        gpst:  


    Returns:
        leapsecs
    """

    import time
    import datetime as dt
    import locale

    # tdiff = (gpsWeek * secsInWeek) + SOW - leapSecs
    # t = t0 + tdiff

    leapsecdict = leapSecDict()
    leapSecs = 0

    if type(dTime) is datetime.datetime:
        pass
    elif (type(dTime) is int) or (type(dTime) is float):
        dTime = dt.datetime.fromtimestamp(dTime)
    elif type(dTime) is tuple:
        t0 = (
            time.mktime(epochTuple) - time.timezone
        )  # mktime is localtime, correct for UTC
        tdiff = dTime[0] * secsInWeek + dTime[1]
        dTime = dt.datetime.fromtimestamp(t0 + tdiff)
    elif dTime is None:
        dTime = dt.datetime.now()

    # print(dTime.strftime("%Y-%b-%d %H:%M:%S"))
    # print(time.mktime( dTime.timetuple()) )

    # making sure locale month format is correct by changing locale.
    syslocale = locale.getlocale()
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")

    for day in leapsecdict.keys():
        if dt.datetime.strptime(day, "%Y-%b-%d") <= dTime:
            leapSecs = leapsecdict[day]

    # changing back to system local.
    locale.setlocale(locale.LC_ALL, "{}.{}".format(*syslocale))

    if gpst is True:

        leapSecs = leapSecs - leapsecdict["1980-Jan-1"]
        # print( "ADDING: {0} s".format(leapSecs) )

    return leapSecs


def leapSecDict():

    leapSecDict = {
        "1972-Jan-1": 10,
        "1972-Jul-1": 11,
        "1973-Jan-1": 12,
        "1974-Jan-1": 13,
        "1975-Jan-1": 14,
        "1976-Jan-1": 15,
        "1977-Jan-1": 16,
        "1978-Jan-1": 17,
        "1979-Jan-1": 18,
        "1980-Jan-1": 19,
        "1981-Jul-1": 20,
        "1982-Jul-1": 21,
        "1983-Jul-1": 22,
        "1985-Jul-1": 23,
        "1988-Jan-1": 24,
        "1990-Jan-1": 25,
        "1991-Jan-1": 26,
        "1992-Jul-1": 27,
        "1993-Jul-1": 28,
        "1994-Jul-1": 29,
        "1996-Jan-1": 30,
        "1997-Jul-1": 31,
        "1999-Jan-1": 32,
        "2006-Jan-1": 33,
        "2009-Jan-1": 34,
        "2012-Jul-1": 35,
        "2015-Jul-1": 36,
        "2017-Jan-1": 37,
    }

    return leapSecDict


# def gpsTimeFromUTC()

# ===== Tests  =========================================


def testTimeStuff():
    print("-" * 20)
    print("The GPS Epoch when everything began (1980, 1, 6, 0, 0, 0, leapSecs=0)")
    (w, sow, d, sod) = gpsFromUTC(1980, 1, 6, 0, 0, 0, leapSecs=0)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow, leapSecs=0))

    print("The time of first Rollover of GPS week (1999, 8, 21, 23, 59, 47)")
    (w, sow, d, sod) = gpsFromUTC(1999, 8, 21, 23, 59, 47)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow, leapSecs=14))

    print("Today is GPS week 1186, day 3, seems to run ok (2002, 10, 2, 12, 6, 13.56)")
    (w, sow, d, sod) = gpsFromUTC(2002, 10, 2, 12, 6, 13.56)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow))


def testJulD():
    print("2002, 10, 11 -> 284  ==??== ", julianDay(2002, 10, 11))


def testGpsWeek():
    print("2002, 10, 11 -> 1187  ==??== ", gpsWeek(2002, 10, 11))


def testDayOfWeek():
    print("2002, 10, 12 -> 6  ==??== ", dayOfWeek(2002, 10, 12))
    print("2002, 10, 6  -> 0  ==??== ", dayOfWeek(2002, 10, 6))


def testPyUtilties():
    ymdhms = (2002, 10, 12, 8, 34, 12.3)
    print("testing for: ", ymdhms)
    pyUtc = apply(mkUTC, ymdhms)
    back = ymdhmsFromPyUTC(pyUtc)
    print("yields     : ", back)
    # *********************** !!!!!!!!
    # assert(ymdhms == back)
    # TODO:: this works only with int seconds!!! fix!!!
    (w, t) = wtFromUTCpy(pyUtc)
    print("week and time: ", (w, t))


# ===== Main =========================================
if __name__ == "__main__":
    pass
    testTimeStuff()
    testGpsWeek()
    testJulD()
    testDayOfWeek()
    testPyUtilties()



# data parsing and calculation/__init__.py

"""Retrieve GPS, Seismological, hydrological and gas data.

Modules exported by this package:

- `gps_read`: Provide several gps processing calculations.
- `sil_read`: Provide several seismological processing calculations, useful to correlate with gps data.
"""




import configparser
import os
import shutil

class ConfigParser:


    def __init__(self):
        # Setting up the working directories
        self.config = configparser.ConfigParser()
        self.mount_path = os.environ.get('MOUNT_PATH')
        current_dir = os.path.dirname(os.path.abspath(__file__))
        
        # Copying the stations.cfg from the mounted path to the current directory
        source_stations_config_path = os.path.join(self.mount_path, 'gpsconfig', 'cparser', 'stations.cfg')
        self.dest_stations_config_path = os.path.join(current_dir, 'stations.cfg')
        shutil.copy(source_stations_config_path, self.dest_stations_config_path)

        # Reading the stations.cfg file
        self.config.read(self.dest_stations_config_path)

        # Copying the postprocess.cfg from the mounted path to the current directory
        source_postprocess_config_path = os.path.join(self.mount_path, 'gpsconfig', 'cparser', 'postprocess.cfg')
        self.dest_postprocess_config_path = os.path.join(current_dir, 'postprocess.cfg')
        shutil.copy(source_postprocess_config_path, self.dest_postprocess_config_path)

        # Reading the postprocess.cfg file
        self.config.read(self.dest_postprocess_config_path)


    # Establishing the methods usable through the package to interact with the cparser module
    def get_config(self, section, option):
        """
        This function gets a configuration option from the 'stations.cfg' file.
        """
        # Getting the configuration option
        value = self.config.get(section, option)
        # Replacing /mnt/ with the MOUNT_PATH
        value = value.replace('/mnt/', self.mount_path)
        return value

    
    def get_stations_config_path(self):
        """
        This function returns the path to the 'stations.cfg' file.
        """
        return self.dest_stations_config_path

    def get_postprocess_config_path(self):
        """
        This function returns the path to the 'postprocess.cfg' file.
        """

        return self.dest_postprocess_config_path

    
    def getStationInfo(self, station_id):
      """
      This function gets station information from the 'stations.cfg' file.
      """
      # Read the 'station' section from the 'stations.cfg' file
      if self.config.has_section(station_id):
          station_info = dict(self.config.items(station_id))
          return {'station': station_info}
      else:
          raise Exception(f"Station '{station_id}' not found in 'stations.cfg' file.")

    def getPostprocessConfig(self, option):
        """
        This function gets a configuration option from the 'postprocess.cfg' file.
        """
        # Read the 'Configs' section from the 'postprocess.cfg' file
        if self.config.has_section('Configs'):
            if self.config.has_option('Configs', option):
                return self.config.get('Configs', option)
        raise Exception(f"Option '{option}' not found in 'Configs' section of the postprocess configuration file.")




"""
This module contains functions for reading and returning GPS data. It includes the following functions 

"""

import numpy as np
import logging
from typing import List, Optional, Union

#
# time series filtering
#

def line(x: Union[int, float], p0: Union[int, float], p1: Union[int, float]) -> Union[int, float]:
    """
    Linear function for detrending timeseries without considering their oscillatory nature.

    Examples:
        >>> line(1, 0, 1)
        1

    Args:
        x (Union[int, float]): Input value in yearf or ordinal (or any numeric time)
        p0 (Union[int, float]): Intercept
        p1 (Union[int, float]): Slope

    Returns:
        Union[int, float]: Output value

    """
    return p0 + p1 * x


def lineperiodic(x: float, p0: float, p1: float, p2: float, p3: float, p4: float, p5: float) -> Optional[float]:
    """
    Linear function with a periodic superimposed

    Examples:
        >>> lineperiodic(1, 0, 1)
        1

    Args:
        x: Input value in yearf or ordinal (or any numeric time)
        p0: Intercept
        p1: Slope of linear term
        p2: Slope of cosine semiannual term
        p3: Slope of sine semiannual term
        p4: Slope of cosine annual term
        p5: Slope of sine annual term

    Returns:
        Output value

    """

    import numpy as np

    return (
        p0
        + p1 * x
        + p2 * np.cos(2 * np.pi * x)
        + p3 * np.sin(2 * np.pi * x)
        + p4 * np.cos(4 * np.pi * x)
        + p5 * np.sin(4 * np.pi * x)
    )


def periodic(x: float, p0: float, p1: float, p2: float, p3: float, p4: float, p5: float)->float:
    """
    Periodic function without the linear part

   
    Examples:
        >>> periodic(1, 0, 1, 1, 1, 1, 1)
        

    Args:
        x: Input value in yearf or ordinal (or any numeric time)
        p0: Coefficient
        p1: Coefficient
        p2: Coefficient
        p3: Coefficient
        p4: Coefficient 
        p5: Coefficient

    Returns:
        Output value

    """

    import numpy as np

    return (
        p2 * np.cos(2 * np.pi * x)
        + p3 * np.sin(2 * np.pi * x)
        + p4 * np.cos(4 * np.pi * x)
        + p5 * np.sin(4 * np.pi * x)
    )


def xf(x: float, p0: float, p1: float, p2: float, tau=4.8 ) -> float:

    """
    Linear + exponential function with constant tau coefficient

    Examples:
        >>> xf(1, 0, 1, 1, 1)

    Args:
        x: Input value in yearf or ordinal (or any numeric time)
        p0: Intercept
        p1: Slope
        p2: Exponential term coefficient
        tau: Exponential coefficient

    Returns:
        y: Output value


    """
    import numpy as np

    return p0 + p1 * x + p2 * np.exp(-tau * x)


def expxf(x: float, p0: float, p1: float, p2: float, p3: float)-> float:
    """

    Linear + exponential function with variable coefficient

    Examples:
        >>> expxf(1, 0, 1, 1, 1)

    Args:
        x: Input value in yearf or ordinal (or any numeric time)
        p0: Intercept
        p1: Slope
        p2: Exponential term coefficient
        p3: Exponential coefficient

    Returns:
        y: Output value

    """

    import numpy as np

    return p0 + p1 * x + p2 * np.exp(-p3 * x)


def expf(x, p0, p1, p2):
    """ 
    Exponential function 

    """
    import numpy as np

    return p0 + p1 * np.exp(-p2 * x)


def gpsvelo_df():
    """
    Dataframe for pygmt velo with three extra columns
    for date, period and boolean for vertical component.
    This is based on the output of the gpsvelo function, which estimates the gps velocity for a specific period of time.

    Examples:
        >>> gpsvelo_df()

    Returns:
        gpsvelo: Dataframe


    """
    import pandas as pd

    gpsvelo = pd.DataFrame(
        columns=[
            "longitude",
            "latitude",
            "east_velo",
            "north_velo",
            "east_sigma",
            "north_sigma",
            "coorelation_EN",
            "Station",
            "date",
            "period",
            "vertical",
        ],
    )

    return gpsvelo


def gpsvelo(sta: str, ll, vel, vertical=False, vfile=None, pheader=False):
    """
    Return gps velocities gmt velo

    Examples:
        >>> gpsvelo()

    Args:
        sta: Station name
        ll: [longitude, latitude]
        vertical: Boolean to describe
        vfile: Velocity file
        pheader: Header

    Returns:
        gpsvelo: Dataframe

    """

    gpsvelo = "{0:5.6f} {1:5.6f}\t{2:7.2f} {5:7.2f}\t{3:7.2f} {6:7.2f}\t{4:7.2f} {7:7.2f}\t\t{8:s}".format(
        ll[1], ll[0], vel[0], vel[1], vel[2], vel[3], vel[4], vel[5], sta
    )


def getDetrFit(
    STA: str,
    useSTA=None,
    useFIT=None,
    onlyPeriodic=False,
    detrFile="detrend_itrf2008.csv",
    logging_level=logging.WARNING):

    """ 
    This function defines the type of detrending to be applied. It is based on the detrending parameters file, detrFile, which is stored in the gpsconfig directory of the gpsplot server.

    Examples:
        >>> getDetrFit('GRIC', useSTA=None, useFIT=None, onlyPeriodic=False, detrFile='detrend_itrf2008.csv')

    Args:
        STA: Station name
        useSTA: Station name
        useFIT: Fit name
        onlyPeriodic: Boolean
        detrFile: Detrending file name, csv format



    """

    import logging
    import numpy as np
    import pandas as pd

    # import cparser as cp
    from cparser import ConfigParser
    
    # Create an instance for the ConfigParser
    config = ConfigParser()

    # Handling logging
    logging.basicConfig(
        format="%(asctime)-15s [%(levelname)s] %(funcName)s: %(message)s",
        level=logging_level,
    )

    logging.getLogger().setLevel(logging_level)
    module_logger = logging.getLogger()

    # Read the detrending file
    North = ["Nrate", "Nacos", "Nasin", "Nscos", "Nssin"]
    East = ["Erate", "Eacos", "Easin", "Escos", "Essin"]
    Up = ["Urate", "Uacos", "Uasin", "Uscos", "Ussin"]

    # Read the detrending file
    #Headers
    Info = ["Sitename", "Starttime", "Endtime", "UseSTA", "Fit"]
    columns = North + East + Up + Info

    # Grab the station name from the station.cfg file
    sta_name = config.getStationInfo(STA)["station"]["station_name"]

    # Initialize the table as an array with nan values
    data = [[np.nan] * 15 + [sta_name, np.nan, np.nan, np.nan, np.nan]]

    # convert to dataframe
    table = pd.DataFrame(data, index=[STA], columns=columns)

    # Read the detrending file
    try:
        table = pd.read_csv(detrFile, index_col="STA")
    except pd.errors.EmptyDataError:
        module_logger.warn("{} does not contain any data".format(detrFile))
    except FileNotFoundError:
        module_logger.warn("Detrend file not found")

    module_logger.info("Existing detrend constant table:\n{}".format(table))

    # Set the fit constants from the file within const variable
    try:
        const = pd.DataFrame(index=[STA], columns=columns)
        const.loc[STA] = tuple(table.loc[STA].values)
    except KeyError:
        module_logger.warn("{} not in table making an empty row".format(STA))
        const = pd.DataFrame(data, index=[STA], columns=table.columns)

    # Set the fit constants from the other station fit
    if useSTA:
        module_logger.warn(
            "Setting {0} {1} constants as {1} parameters for {2}".format(
                useSTA, useFIT, STA
            )
        )
        module_logger.info("current {} constants are:\n {} ".format(STA, const))

        # Set the fit constants if the fit is done from another station
        const.loc[STA, ["UseSTA", "Fit"]] = useSTA, useFIT

        # Set the fit constants
        if useFIT == "periodic":
            const.loc[STA, North[1:] + East[1:] + Up[1:]] = table.loc[
                useSTA, North[1:] + East[1:] + Up[1:]
            ]
        elif useFIT == "lineperiodic":
            const.loc[STA, North + East + Up] = table.loc[
                useSTA, North + East + Up
            ]
        elif useFIT == "line":
            const.loc[STA, North[:1] + East[:1] + Up[:1]] = table.loc[
                useSTA, North[:1] + East[:1] + Up[:1]
            ]

        module_logger.info("New parameters for {} are:\n {}".format(STA, const))

    # Set the linear part of the fit constants as nan for periodic only fit
    if onlyPeriodic is True:
        const["Nrate"] = const["Erate"] = const["Urate"] = np.nan

    return const

def convconst(const, pb=None):
    """
    This function converts detrending constants from pandas dataframe to list of the form
    [[], [], []] for optimize.curv_fit

    Examples:
        >>> convconst()

    Args:
        const: detrend constants dataframe
        pb: Periodic constants

    Returns:
        const: detrending constants

    """


    import numpy as np

    North = ["Nrate", "Nacos", "Nasin", "Nscos", "Nssin"]
    East = ["Erate", "Eacos", "Easin", "Escos", "Essin"]
    Up = ["Urate", "Uacos", "Uasin", "Uscos", "Ussin"]

    if pb:
        # const = pd.DataFrame(data, index=[STA], columns=table.columns)
        stat = const.index.tolist()[0]
        for i, dimention in zip(range(3), [North, East, Up]):
            const.loc[stat, dimention] = list(pb[i][1:])

        return const

    else:
        if const[North].isna().sum(axis=1).values == 5:
            p0 = [None, None, None]
        else:
            p0 = [[], [], []]
            p0[0] = [0] + const[North].values.tolist()[0]
            p0[1] = [0] + const[East].values.tolist()[0]
            p0[2] = [0] + const[Up].values.tolist()[0]

            p0 = np.nan_to_num(p0)

        return p0


def save_detrend_const(const, detrFile="detrend_itrf2008.csv"):
    """
    This function saves constants in a file for detrending GPS stations

    Examples:
        >>> save_detrend_const(const, detrFile='detrend_itrf2008.csv')

    Args:
        const: detrending constants dataframe
        detrFile: detrending file name

    Returns:
        None


    """
    import pandas as pd
    from pathlib import Path

    North = ["Nrate", "Nacos", "Nasin", "Nscos", "Nssin"]
    East = ["Erate", "Eacos", "Easin", "Escos", "Essin"]
    Up = ["Urate", "Uacos", "Uasin", "Uscos", "Ussin"]
    Info = ["Sitename", "Starttime", "Endtime", "UseSTA", "Fit"]

    columns = North + East + Up + Info

    stationlist = const.index.values
    path = Path(Path.cwd(), detrFile)
    if not path.is_file():
        open(path, "w").close()

    try:
        table = pd.read_csv(detrFile, index_col="STA")
    except pd.errors.EmptyDataError:
        table = pd.DataFrame(index=["STA"], columns=columns)

    if all(table.isnull().all(1)):
        print("No data in file")
        const.to_csv(detrFile, mode="w", header=True, index_label="STA")

        return

    else:
        for stat in stationlist:
            if stat in table.index:
                print("Station {} already in table updating".format(stat))
                table.loc[stat, :] = const.loc[stat]
                print("Saving the new table")
                table.to_csv(detrFile, mode="w", header=True, index_label="STA")
            else:
                print("stations {} are not in detrend file, adding it".format(stat))
                table.loc[stat, :] = const.loc[stat]
                table.to_csv(detrFile, mode="w", header=True, index_label="STA")

        return


def fitdataframe(func, df, p0=[None, None, None]):
    """ 
    This function finds the fitting parameters for a dataframe of GPS time series through the scipy.optimize.curve_fit function

    Examples:
        >>> fitdataframe(line, df, p0=[None, None, None])

    Args:
        func: fitting function (line, periodic, lineperiodic, others)
        df: dataframe of GPS time series, with x in fractional year, y the components dN, dE or dU and yD, the uncertainties DN, DE, DU.
        p0: initial guess for the fitting parameters

    Returns:
        pb: fitting parameters
        pcov: covariance matrix
        
    """
    from scipy import optimize

    x = df["yearf"].to_numpy()
    y = df[["north", "east", "up"]].to_numpy().T
    yD = df[["Dnorth", "Deast", "Dup"]].to_numpy().T

    return fittimes(func, x, y, yD, p0=p0)


def fittimes(func, x, y, yD=[None, None, None], p0=[None, None, None]):

    from scipy import optimize

    pb = [[], [], []]
    pcov = [[], [], []]

    for i in range(3):
        pb[i], pcov[i] = optimize.curve_fit(
            func, x, y[i], p0=p0[i], sigma=yD[i], maxfev=100000
        )

    return pb, pcov


# routines to extract and save coordinates and time series from gamit


def gamittooneuf(sta, outFile, mm=True, ref="plate", dstring=None, outformat=True):
    """
    extract Gamit timeseries from standard format to one file formated time string

    Examples:
        >>> gamittooneuf('GRIC', outFile, mm=True, ref='plate', dstring=None, outformat=True)

    Args:
        sta: station four letter short name
        outFile: file object
        mm: boolean True return data in mm, else in m
        ref: subtract plate velocity from the time series.
        dstring: format of the time string (e.g dstring)="%Y%m%d-%H%M%S"), defaults to decimal year (yyyy.yyyy)


    Returns: #not finished but determines the output order of the data into the file
        .outformat file

    """
    # "%Y/%m/%d 12:00:00.000"

    NEUdata = gamittoNEU(sta, mm=mm, ref=ref, dstring=dstring)

    gamittoFile(NEUdata, outFile, mm=mm, ref=ref, dstring=dstring, outformat=outformat)


def gamittoFile(NEUdata, outFile, mm=True, ref="plate", dstring=None, outformat=True):
    """ 
    This function converts gamit timeseries files to NEU files

    Examples:
        >>> gamittoFile(NEUdata, outFile, mm=True, ref='plate', dstring=None, outformat=True)

    Args:
        NEUdata: NEU data
        outFile: file object
        mm: boolean True return data in mm, else in m
        ref: subtract plate velocity from the time series if = ref.
        dstring: format of the time string (e.g dstring)="%Y%m%d-%H%M%S"), defaults to decimal year (yyyy.yyyy)
        outformat: file format (True) or string format (False)

    Returns:
        .outformat file


    """

    # formatting time column
    if dstring == "yearf":
        timef = "{0: 8.4f}\t"
        timeh = "#\"yyyy.dddd'     "
    else:
        timef = " {0:s}\t"
        timeh = "#yyyy/mm/dd HH:MM:SS.SSS          "

    # unit conversion formatting
    if mm:
        if outformat:
            header = timeh + "dN[mm] DN[mm]\tdE[mm] DE[mm]\tdU[mm]  DU[mm]"
            formatstr = (
                timef + "{1: 7.2f} {4: 7.2f}\t{2: 7.2f} {5: 7.2f}\t{3: 7.2f} {6: 7.2f}"
            )
        else:
            header = timeh + "dN[mm]  dE[mm] dU[mm]\t\t  DN[mm]  DE[mm]  DU[mm]"
            formatstr = (
                timef
                + "{1: 7.2f} {2: 7.2f} {3: 7.2f}\t\t{4: 7.2f} {5: 7.2f}\t{6: 7.2f}"
            )
    else:
        if outformat:
            header = timeh + "dN[m] DN[m]\tdE[m] DE[m]\tdU[m]  DU[m]"
            formatstr = (
                timef + "{1: 7.5f} {4: 7.5f}\t{2: 7.5f} {5: 7.5f}\t{3: 7.5f} {6: 7.5f}"
            )
        else:
            header = timeh + "dN[m]    dE[m]    dU[m]          DN[m]   DE[m]    DU[m]"
            formatstr = (
                timef + "{1: 7.5f} {2: 7.5f} {3: 7.5f}\t{4: 7.5f} {5: 7.5f} {6: 7.5f}"
            )

    print('printing header and file \n', header, file=outFile)  

    # applying formatting to each column of the NEUdata file
    for x in NEUdata:
        print(
            formatstr.format(
                x["yearf"],
                x["data[0]"],
                x["data[1]"],
                x["data[2]"],
                x["Ddata[0]"],
                x["Ddata[1]"],
                x["Ddata[2]"],
            ),
            file=outFile,
        )


def savedisp(dataDict, fname=None, header=""):
    """ 
    This function saves a dictionary of data to a file

    Examples:
        >>> savedisp(dataDict, fname=None, header="")

    Args:
        dataDict: dictionary of data
        fname: file name
        header: header of the file

    Returns:
        output file

    """

    from collections import OrderedDict

    valtype = type(dataDict.values()[0])

    dataDict = OrderedDict(sorted(dataDict.items()))

    if (valtype is list) or (valtype is np.ndarray):
        fmt = "% 3.8f\t% 2.8f\t% 2.8f\t%s" # format

        
        ab = np.zeros(
            len(dataDict.keys()),
            dtype=[
                ("var1", "float"),
                ("var2", "float"),
                ("var3", "float"),
                ("var4", "a4"),
            ],
        )

        ab["var1"] = np.squeeze(dataDict.values())[:, 0]
        ab["var2"] = np.squeeze(dataDict.values())[:, 1]
        ab["var3"] = np.squeeze(dataDict.values())[:, 2]
        ab["var4"] = dataDict.keys()


    # formatting tuples into a dictionary of arrays
    if valtype is tuple:
        fmt = "% 3.8f\t% 2.8f\t% 2.8f\t%2.8f\t%2.8f\t%s"
        ab = np.zeros(
            len(dataDict.keys()),
            dtype=[
                ("var1", "float"),
                ("var2", "float"),
                ("var3", "float"),
                ("var4", "float"),
                ("var5", "float"),
                ("var6", "a4"),
            ],
        )
        ab["var1"] = np.squeeze(zip(*dataDict.values()[:])[0])[:, 0]
        ab["var2"] = np.squeeze(zip(*dataDict.values()[:])[0])[:, 1]
        ab["var3"] = np.squeeze(zip(*dataDict.values()[:])[1])[:, 0]
        ab["var4"] = np.squeeze(zip(*dataDict.values()[:])[1])[:, 1]
        ab["var5"] = np.squeeze(zip(*dataDict.values()[:])[1])[:, 2]
        ab["var6"] = dataDict.keys()

    if fname:
        np.savetxt(fname, ab, fmt=fmt, header=header)
    return ab


def extractfromGamitBakf(cfile, stations):
    """
    Function to extract data from a GAMIT .bak file

    Examples:
        >>> extractfromGamitBakf(cfile, stations)

    Args:
        cfile: .bak file name
        stations: station name

    Returns:
        slines: list of lines
    """
    import re

    slines = []

    site = re.compile(stations)
    tim = re.compile("Solution refers to", re.IGNORECASE)
    f = open(cfile, "r")

    for line in f:
        if site.search(line):  # or tim.search(line):
            slines.append(line.rstrip())

    return slines


def openGlobkTimes(STA, Dir=None, tType="TOT"):
    """
    Function to import data from Globk time series files into a numpy arrays

    Dir is the directory containing the time series if left blank the default path will be the path
    defined in the config file postprossesing.cfg, totPath

    Args:
        STA: Station four letter short name in captial letters
        Dir: optional alternative Directory of the GAMIT time series data.

    Returns:
    
        yearf: is array with time  (usually) in fractional year format (i.e. 2014.62328)
        data:  three arrays containing GPS data in north,east, up
        Ddata: respective uncertainty values

    """

    import os
    import numpy as np
    import datetime as dt
    from gtimes.timefunc import shifTime, TimetoYearf
    from cparser import ConfigParser
    config = ConfigParser()



    # Loading the data 
    # First, if no Dir is input, find the path in the Postprocess config file in the "totpath" instance
    if Dir is None:
        Dir = config.getPostprocessConfig("totpath")

    else: # This is still in development
        # print( os.path.isdir(Dir) )
        # check if the path given as argument indeed exists
        pass

    # constructing the full path filenames and parsing parameters
    # define the file name to be looked for into Dir
    filepre = "mb_{0:s}_{1:s}.dat".format(STA, tType)

    if os.path.isfile(os.path.join(Dir, filepre + "1")): #use dat1, dat2, dat3 as file format
        pass
    else: # use TOT as file format to read in data
        filepre = "mb_{0:s}_{1:s}.dat".format(STA, "TOT")

    # Construct the file names for each component
    Datafile1 = os.path.join(Dir, filepre + "1")
    Datafile2 = os.path.join(Dir, filepre + "2")
    Datafile3 = os.path.join(Dir, filepre + "3")

    # Load the data for files and store it in arrays. Use converters defined in the __converter functions
    yearf, d1, D1 = np.loadtxt(
        Datafile1, unpack=True, skiprows=3, converters={1: __converter, 2: __converter}
    )
    d2, D2 = np.loadtxt(
        Datafile2,
        usecols=(1, 2),
        unpack=True,
        skiprows=3,
        converters={1: __converter, 2: __converter},
    )
    d3, D3 = np.loadtxt(
        Datafile3,
        usecols=(1, 2),
        unpack=True,
        skiprows=3,
        converters={1: __converter, 2: __converter},
    )

    # Stack the arrays for the three components in the data array and for their uncertainties in Ddata array
    data = np.vstack([d1, d2, d3])
    Ddata = np.vstack([D1, D2, D3])

    # Option to grab 8hr subdaily solutions from the same path
    if tType == "08h":
        shift8h = dt.timedelta(**shifTime("H8"))
        yearf = np.array(
            [
                TimetoYearf(*(item + shift8h).timetuple()[:6])
                for item in toDateTime(yearf)
            ]
        )

    return yearf, data, Ddata


def open3DataFiles(STA, Dir=None, comp=["-N", "-E", "-U"]):
    """
    This function opens data contained in 3 files, one file for each component E, N and U

    Examples:
        >>> open3DataFiles(STA, Dir=None, comp=["-N", "-E", "-U"])

    Args:
        STA: Station four letter short name in capital letters
        Dir: optional alternative Directory of the files.
        comp: list of components to be read. Default is ["-N", "-E", "-U"]
        
    Returns:
        data:  three arrays containing GPS data in north, east and up

    """

    import os
    import pandas as pd

    if Dir is None:
        Dir = os.getcwd()

    compdict = {}
    compdict[comp[0]] = ["north", "Dnorth"]
    compdict[comp[1]] = ["east", "Deast"]
    compdict[comp[2]] = ["up", "Dup"]
    components = {"north": None, "east": None, "up": None}

    for item in compdict.keys():
        dfile = "{0}{1}".format(STA, item)
        components[compdict[item][0]] = pd.read_csv(
            dfile, sep=r"\s+", index_col=0, header=None, names=compdict[item]
        )

    columnreorder = ["north", "east", "up", "Dnorth", "Deast", "Dup"]
    data = pd.concat(components.values(), axis=1)[columnreorder]
    data.set_index(pd.DatetimeIndex(toDateTime(data.index)), inplace=True)
    data.index = data.index.round("1h")

    return data


def convGlobktopandas(yearf: float, data, Ddata):
    """
    This function converts Globk files (from OpenGlobkTimes function) into pandas dataframes

    Examples:
        >>> convGlobktopandas(yearf, data, Ddata)

    Args:
        yearf: is array with time  (usually) in fractional year format (i.e. 2014.62328)
        data:  three arrays containing GPS data in north, east and up
        Ddata: respective uncertainty values for each component 

    Returns:
           Index [datetime], north, east, up, Dnorth, Deast, Dup, yearf


    """

    import pandas as pd
    from collections import OrderedDict

    # reduce(lambda x, y: pd.merge(x, y, on = 'Date'), dfList)
    names = ["north", "east", "up", "Dnorth", "Deast", "Dup", "yearf"]

    # !!!!! from_item depricated will remove
    # data = pd.DataFrame.from_items(zip(names[:3],data))
    # data = data.join( pd.DataFrame.from_items(zip(names[3:],Ddata) ) )
    # Using from_dict instead
    data = pd.DataFrame.from_dict(OrderedDict(zip(names[:3], data)))
    data = data.join(pd.DataFrame.from_dict(OrderedDict(zip(names[3:], Ddata))))
    data["yearf"] = yearf
    data.set_index(pd.DatetimeIndex(toDateTime(yearf)), inplace=True)
    data.index = data.index.round("1h")

    return data


def compGlobkTimes(stalist="any", dirConFilePath=None, freq=None):
    """
    This function joins old and new mb_ time series files

    Examples:
        >>> compGlobkTimes(stalist="any", dirConFilePath=None, freq=None)

    Args:
        stalist: list of station names in capital letters. If "any", all stations are used. Default is "any"
        dirConFilePath: optional alternative Directory of the GAMIT time series data.
        freq: optional frequency of the data. Default is None

    Returns:
        data:  three arrays containing GPS data in north, east and up
    
    """
    import glob
    import sys
    import shutil
    import os
    from cparser import ConfigParser

    config = ConfigParser()

    #totpath = config.getPostprocessConfig('totpath')
    #import cparser as cp

    if dirConFilePath:  # for custom file
        Dirs = parsedir(dirConFilePath)
    else: # grab paths from the postprocess.cfg file in gpsconfig directory that cparser reads into a dictionary
        Dirs = {
                "figDir": config.get_config("Configs", "figDir"),
                "prePath": config.get_config("Configs", "prePath"),
                "rapPath": config.get_config("Configs", "rapPath"),
                "totPath": config.get_config("Configs", "totPath")
            }

        print('Directory is \n', Dirs)

    # Reading into a string the paths
    PrePath = Dirs["prePath"]
    RapPath = Dirs["rapPath"]
    TotPath = Dirs["totPath"]

    # Setting the frequency to TOT if freq is None
    if freq == "TOT" or freq is None:
        freq = "TOT"
    else: # setting the frequency
        PrePath = PrePath + "_%s" % (freq)
        RapPath = RapPath + "_%s" % (freq)

    if stalist == "any":
        FilePreL = os.path.join(PrePath, "mb_*.dat?")
        FileRapL = os.path.join(RapPath, "mb_*.dat?")

        List = glob.glob(FilePreL) + glob.glob(FileRapL)

        # listing all stations in  the Rap and Pre directories
        stalist = sorted(set([item[-13:-9] for item in List]))

    # Set up names for files
    for STA in stalist:
        FilePre = "mb_%s_?PS.dat" % STA
        OutFilePre = "mb_%s_GPS.dat" % STA
        GPS20PS = "mb_%s_0PS.dat" % STA

        for axes in range(1, 4):
            FilePreR = os.path.join(PrePath, FilePre + "%s" % (axes,))
            FileRapR = os.path.join(RapPath, FilePre + "%s" % (axes,))

            # graping the list for files for for that station
            PreFileL = glob.glob(FilePreR)  # listing files in the pre dir
            RapFileL = glob.glob(FileRapR)  # listing files in th Rap dir

            #  Sorting the file lists
            PreFileL.sort()
            if len(PreFileL) > 1:
                PreFileL.insert(0, PreFileL.pop(-1))
            RapFileL.sort()
            if len(RapFileL) > 1:
                RapFileL.insert(0, RapFileL.pop(-1))

            TotFile = os.path.join(TotPath, "mb_%s_%s.dat%s" % (STA, freq, axes))

            print("Concatenating all the %s data to %s" % (STA, TotFile))

            if os.path.exists(TotFile):
                os.remove(TotFile)

            outf = open(TotFile, "a")

            for fil in PreFileL:
                print("Processing file %s " % fil, file=sys.stderr)
                f = open(fil)
                f.seek(61)
                shutil.copyfileobj(f, outf)
                f.close()

            outf.close()

            preexist = os.stat(TotFile).st_size != 0
            if preexist:
                outf = open(TotFile, "r")
                lastline = outf.readlines()[-1]
                lastline = lastline.split()
                outf.close()

            outf = open(TotFile, "a")
            for file in RapFileL:
                formatstr = "Processing file {0:s} ".format(file)
                print(formatstr, file=sys.stderr)
                rapfile = open(file, "r")
                rapfile.seek(61)
                lines = rapfile.readlines()
                if preexist:
                    lines = "".join(
                        [line for line in lines if line.split()[0] > lastline[0]]
                    )
                else:
                    lines = "".join([line for line in lines])

            outf.close()


def TieTimes(sta1, sta2, dirConFilePath=None, freq=None, tie=[None, None, None]):
    """
    This function joins old and new mb_ time series files

    Examples:
        >>> TieTimes(sta1, sta2, dirConFilePath=None, freq=None, tie=[None, None, None])

    Args:
        sta1: first station name in capital letters
        sta2: second station name in capital letters
        dirConFilePath: optional alternative Directory of the GAMIT time series data.
        freq: optional frequency of the data. Default is None
        tie: optional tie file. Default is [None, None, None]


    """
    import glob
    import sys
    import shutil
    from cparser import ConfigParser
    config = ConfigParser()
    #totpath = config.getPostprocessConfig('totpath')

    if dirConFilePath:  # for custom file
        Dirs = parsedir(dirConFilePath)
    else:
        # As the standard configparser works with dictionaries, use it to create the Dirs dictionaries
        Dirs = {
        "figDir": config.get_config("Configs", "figDir"),
        "prePath": config.get_config("Configs", "prePath"),
        "rapPath": config.get_config("Configs", "rapPath"),
        "totPath": config.get_config("Configs", "totPath")
    }

    # parse the paths from the Dirs dictionary
    PrePath = Dirs["prePath"]
    RapPath = Dirs["rapPath"]
    TotPath = Dirs["totPath"]

    # Setting the frequency to TOT if freq is None
    if freq == "TOT" or freq is None:
        freq = "TOT"
    else:
        PrePath = PrePath + "_%s" % (freq)
        RapPath = RapPath + "_%s" % (freq)

    # for all the stations, create the path for PreL and RapL files
    if stalist == "any":
        FilePreL = os.path.join(PrePath, "mb_*.dat?")
        FileRapL = os.path.join(RapPath, "mb_*.dat?")

        List = glob.glob(FilePreL) + glob.glob(FileRapL)

        # listing all stations in  the Rap and Pre dir
        stalist = sorted(set([item[-13:-9] for item in List]))

    for STA in stalist:
        FilePre = "mb_%s_?PS.dat" % STA
        OutFilePre = "mb_%s_GPS.dat" % STA
        GPS20PS = "mb_%s_0PS.dat" % STA

        for axes in range(1, 4):
            FilePreR = os.path.join(PrePath, FilePre + "%s" % (axes,))
            FileRapR = os.path.join(RapPath, FilePre + "%s" % (axes,))

            # graping the list for files for for that station
            PreFileL = glob.glob(FilePreR)  # listing files in the pre dir
            RapFileL = glob.glob(FileRapR)  # listing files in th Rap dir

            #  Sorting the file lists
            PreFileL.sort()
            if len(PreFileL) > 1:
                PreFileL.insert(0, PreFileL.pop(-1))
            RapFileL.sort()
            if len(RapFileL) > 1:
                RapFileL.insert(0, RapFileL.pop(-1))

            TotFile = os.path.join(TotPath, "mb_%s_%s.dat%s" % (STA, freq, axes))
            print("Concating all the %s data to %s" % (STA, TotFile))
            if os.path.exists(TotFile):
                os.remove(TotFile)
            outf = open(TotFile, "a")
            for fil in PreFileL:
                print("Processing file %s " % fil, file=sys.stderr)
                f = open(fil)
                f.seek(61)
                shutil.copyfileobj(f, outf)
                f.close()
            outf.close()

            preexist = os.stat(TotFile).st_size != 0
            if preexist:
                outf = open(TotFile, "r")
                lastline = outf.readlines()[-1]
                lastline = lastline.split()
                outf.close()

            outf = open(TotFile, "a")
            for file in RapFileL:
                formatstr = "Processing file {0:s} ".format(file)
                print(formatstr, file=sys.stderr)
                rapfile = open(file, "r")
                rapfile.seek(61)
                lines = rapfile.readlines()
                if preexist:
                    lines = "".join(
                        [line for line in lines if line.split()[0] > lastline[0]]
                    )
                else:
                    lines = "".join([line for line in lines])

                outf.write(lines)
                rapfile.close()

            outf.close()


def TieTimes(sta1, sta2, dirConFilePath=None, freq=None, tie=[None, None, None]):
    """
    This function joins old and new mb_ time series files

    Examples:
        >>> TieTimes(sta1, sta2, dirConFilePath=None, freq=None, tie=[None, None, None])

    Args:
        sta1: first station name in capital letters
        sta2: second station name in capital letters
        dirConFilePath: optional alternative Directory of the GAMIT time series data.
        freq: optional frequency of the data. Default is None
        tie: optional tie file. Default is [None, None, None]

    Returns:
        None


    """
    import glob
    import sys
    import shutil

    from pandas import read_table
    import pandas as pd
    from cparser import ConfigParser
    config = ConfigParser()

    if dirConFilePath:  # for custom file
        Dirs = parsedir(dirConFilePath)
    else:
        Dirs = {
                "figDir": config.get_config("Configs", "figDir"),
                "prePath": config.get_config("Configs", "prePath"),
                "rapPath": config.get_config("Configs", "rapPath"),
                "totPath": config.get_config("Configs", "totPath")
            }


    # PrePath = Dirs['prePath'] - These paths are not used for now, but can be added later
    # RapPath = Dirs['rapPath'] - Same case as above
    TieFile = Dirs["tiefile"]
    TotPath = Dirs["totPath"]

    if freq == "TOT" or freq is None:
        freq = "TOT"
    else:
        PrePath = PrePath + "_%s" % (freq)
        RapPath = RapPath + "_%s" % (freq)

    print(TieFile)

    dtype = [
        ("North", "<f8"),
        ("East", "<f8"),
        ("Up", "<f8"),
        ("sta1", "|S5"),
        ("sta2", "|S5"),
    ]

    const = np.genfromtxt(TieFile, dtype=dtype)
    const = [i for i in const if i[3] == sta1 and i[4] == sta2]
    print(const)

    for axes in range(1, 4):
        TotFile1 = os.path.join(TotPath, "mb_%s_%s.dat%s" % (sta1, freq, axes))
        TotFile2 = os.path.join(TotPath, "mb_%s_%s.dat%s" % (sta2, freq, axes))
        print("Concating all the %s data to %s" % (sta1, TotFile2))
        # outf = open(TotFile, 'r')
        data1 = read_table(
            TotFile1, sep="\s+", header=None, index_col=0, names=["disp", "uncert"]
        )
        data2 = pd.read_csv(
            TotFile2, sep="\s+", header=None, index_col=0, names=["disp", "uncert"]
        )
        print(const[0][axes - 1])
        data2["disp"] -= const[0][axes - 1] / 1000
        data = pd.concat([data1, data2])
        

        outfile = os.path.join(TotPath, "mb_%s_%s.dat%s" % (sta2, "JON", axes))
        data.to_csv(outfile, sep="\t", index=True, header=False)


fitfuncl = lambda p, x: p[0] * x + p[1]
errfuncl = lambda p, x, y: fitfuncl(p, x) - y  # Distance to the target function

fitfunc = (
    lambda p, x: p[0] * x
    + p[1] * np.cos(2 * np.pi * x)
    + p[2] * np.sin(2 * np.pi * x)
    + p[3] * np.cos(4 * np.pi * x)
    + p[4] * np.sin(4 * np.pi * x)
    + p[5]
)
errfunc = lambda p, x, y: fitfunc(p, x) - y  # Distance to the target function


def fitline(yearf, data, STA):
    """
    This function fits a function through data points of a station STA

    Examples:
        >>> fitline(yearf, data, STA)

    Args:
        yearf: list of years
        data: list of data
        STA: station name

    Returns:
        parameters of the linear fitting


    """

    from scipy import optimize
    import numpy as np

    dtype = [
        ("Nrate", "<f8"),
        ("Erate", "<f8"),
        ("Urate", "<f8"),
        ("Nacos", "<f8"),
        ("Nasin", "<f8"),
        ("Eacos", "<f8"),
        ("Easin", "<f8"),
        ("Uacos", "<f8"),
        ("Uasin", "<f8"),
        ("Nscos", "<f8"),
        ("Nssin", "<f8"),
        ("Escos", "<f8"),
        ("Essin", "<f8"),
        ("Uscos", "<f8"),
        ("Ussin", "<f8"),
        ("shortname", "|S5"),
        ("name", "|S20"),
    ]

    const = np.genfromtxt("itrf08det", dtype=dtype)
    const = [i for i in const if i[15] == STA]

    pN = [const[0][0]]
    pE = [const[0][1]]
    pU = [const[0][2]]
    pN = [-1 * i for i in pN]
    pE = [-1 * i for i in pE]
    pU = [-1 * i for i in pU]
    # pN.append(0)
    # pE.append(0)
    # pU.append(0)

    # print "pN: %s" % p
    # print "pE: %s" % pE
    # print "pU: %s" % pU

    pb = [[0, 0], [0, 0], [0, 0]]

    # pb[0], success = optimize.leastsq(errfunc, pN[:], args=(yearf-yearf[0], data[0]))
    # pb[1], success = optimize.leastsq(errfunc, pE[:], args=(yearf-yearf[0], data[1]))
    # pb[2], success = optimize.leastsq(errfunc, pU[:], args=(yearf-yearf[0], data[2]))
    pb[0], success = optimize.leastsq(errfuncl, pb[0], args=(yearf, data[0]))
    pb[1], success = optimize.leastsq(errfuncl, pb[1], args=(yearf, data[1]))
    pb[2], success = optimize.leastsq(errfuncl, pb[2], args=(yearf, data[2]))

    return pN, pE, pU, pb


def pvel(pl, pcov):
    """ 
    This function prints the velocity for the three components and their uncertainty to numpy arrays

    Examples:
        >>> pvel(pl, pcov)

    Args:
        pl: list of the three components of the velocity and their uncertainties
        pcov: covariance matrix of the three components

    Returns:
        vel: list of the three components of the velocity
        vunc: list of the three components of the uncertainties
    """

    vunc = [None, None, None]
    vel = [None, None, None]

    for i in range(3):
        vel[i] = pl[i][1]
        vunc[i] = np.sqrt(np.diag(pcov[i]))[1]
        # print("{0:0.2f} {1:0.2f}".format( pl[i][1],vunc[i]) )

    return vel, vunc


def printvelocity(sta, ll, vel, vfile, pheader=False):
    """
    This function prints the velocity for a station to a file

    Examples:
        >>> printvelocity('GRIC', ll, vel, vfile)

    Args:
        sta: station name
        ll: list of the longitude and latitude
        vel: list of the three components of the velocity
        vfile: file to print the velocity
        pheader: print the header

    Returns:
        gpsvelo: string of the velocity

    """
    import os

    header = (
        "#lon       lat\t\t   N[mm]   DN[mm]  E[mm]  DE[mm]  U[mm]  DU[mm]\t\tStation"
    )

    if pheader is True:
        print(header, file=vfile)

    gpsvelo = "{0:5.6f} {1:5.6f}\t{2:7.2f} {5:7.2f}\t{3:7.2f} {6:7.2f}\t{4:7.2f} {7:7.2f}\t\t{8:s}".format(
        ll[1], ll[0], vel[0], vel[1], vel[2], vel[3], vel[4], vel[5], sta
    )

    print(gpsvelo, file=vfile)

    return gpsvelo


def detrend(
    x,
    y,
    Dy=None,
    fitfunc=lineperiodic,
    p=None,
    pcov=None,
    STA=None,
    onlyPeriodic=True,
    zref=False,
):
    """
    This function returns the detrending parameters for a specific fit function and station, for the GPS components

    Examples:
        >>> detrend(x, y, Dy, fitfunc=lineperiodic, p=None, pcov=None, STA='GRIC', onlyPeriodic=True, zref=False)

    Args:
        x: list of the years fractional for the timeseries
        y: list of the three components measurements for the timeseries
        Dy: list of the uncertainties of the three components
        fitfunc: fit function for the detrending. Default is lineperiodic
        p: list with initial guess for the fitting coefficients. Default is None.
        pcov: covariance matrix of the initial guess. Default is None.
        STA: station name, four letters in upper case. Default is None.
        onlyPeriodic: True or False. Default is True.
        zref: Reference z. Default is False. - Not implemented fully yet.

    Returns:
        


    """

    import numpy as np
    import geo_dataread.gps_read as gdrgps

    if Dy is None:
        Dy = np.ones(y.shape)

    # Handling parameters
    if p is not None:
        pass
    else:
        if STA:
            const = getDetrFit(STA, onlyPeriodic=onlyPeriodic)
            p0 = convconst(const)
        else:
            p0 = [None, None, None]

        p, pcov = fittimes(fitfunc, x, y, Dy, p0=p0)


    for i in range(3):
        y[i] = y[i] - fitfunc(x, *p[i])

    if zref:
        _, y, _ = gdrgps.vshift(x, y, Dy, uncert=20.0, refdate=None, Period=5)

    return y


def dPeriod(yearf, data, Ddata, startyear=None, endyear=None):
    """
    update( dict( [ [ line.split(',')[0], line.split(',')[1:] ] for line in args.eventf.read().splitlines() ] ) )

    This function returns the data for a specific period. Its default behavior is to do nothing.

    Examples:
        >>> dPeriod(yearf, data, Ddata, startyear=None, endyear=None)

    Args:
        yearf: time array with numeric time values
        data: data array 
        Ddata: same form as data but containing the uncertainties
        startyear: Initial time. Default=None
        endyear: Final time. Default=None

    Returns:
        yearf: Time array within the period defined by startyear and endyear
        data:  Data within the period defined by startyear and endyear
        Ddata: Data within the period defined by startyear and endyear

    """
    if startyear:
        index = np.where(yearf <= startyear - 0.001)
        yearf = np.delete(yearf, index)
        data = np.delete(data, index, 1)
        Ddata = np.delete(Ddata, index, 1)

    if endyear:
        index = np.where(yearf >= endyear + 0.001)
        yearf = np.delete(yearf, index)
        data = np.delete(data, index, 1)
        Ddata = np.delete(Ddata, index, 1)

    return yearf, data, Ddata


def vshift(yearf, data, Ddata, uncert=20.0, refdate=None, Period=5, offset=None):
    """
    This function shifts time series data by the average value of the interval defined by reday and the number of days specified ()

    Examples:
        >>> vshift(yearf, data, Ddata, uncert=20.0, refdate=None, Period=5, offset=None)

    Args:
        yearf: time array with numeric time values
        data: data array
        Ddata: same form as data but containing the uncertainties
        uncert: Maximum uncertainty of the data. Default=20.0
        refdate: reference date. Default=None
        Period: number of days. Default=5
        offset: initial offset. Default=None

    Returns:
        yearf: Time array
        data:  Data
        Ddata: Data uncertainty

    """

    from gtimes.timefunc import currYearfDate

    # Filtering a little, removing big outliers
    with np.errstate(invalid="ignore"):
        filt = Ddata < uncert
    filt = np.logical_and(np.logical_and(filt[0, :], filt[1, :]), filt[2, :])

    yearf = yearf[filt]
    data = np.reshape(data[np.array([filt, filt, filt])], (3, -1))
    Ddata = np.reshape(Ddata[np.array([filt, filt, filt])], (3, -1))

    if data.any():
        if not (offset is None):
            pass
        else:
            offset = estimate_offset(yearf, data, Ddata, refdate=refdate, Period=Period)

    data = np.array([data[i, :] - offset[i] for i in range(3)])

    return yearf, data, Ddata, offset


def estimate_offset(yearf, data, Ddata, refdate=None, Period=5):
    """
    Estimating offset of a time series at a reference (refdate) point for a given interval (Period)
    defaults at 5 days at the start of the time series

    """
    # averaging the first period days
    if refdate:
        startdate = currYearfDate(0, refdate)
        enddate = currYearfDate(Period, refdate)
        if Period < 0:
            tmpyearf, tmpdata, tmpDdata = dPeriod(
                yearf, data, Ddata, enddate, startdate
            )
        else:
            tmpyearf, tmpdata, tmpDdata = dPeriod(
                yearf, data, Ddata, startdate, enddate
            )

        if tmpdata.any():
            # if there are any data from this period
            offset = np.average(tmpdata[0:3, :], 1, weights=1 / tmpDdata[0:3, :])
        else:
            # We need to extrapolate
            # þarf að díla við þetta með því að módelera.
            offset = np.average(data[0:3, 0:j], 1, weights=1 / Ddata[0:3, 0:7])
    else:
        offset = np.average(data[0:3, 0:Period], 1, weights=1 / Ddata[0:3, 0:Period])

    return offset


def iprep(yearf, data, Ddata, uncert=20.0, offset=None):
    """
    This function is a wrapper for vshift intendet for initializing the time series. It converts to mm and initializes the start of the time series to zero.

    Examples:
        >>> iprep(yearf, data, Ddata, uncert=20.0, offset=None)

    Args:
        yearf: time array with numeric time values
        data: data array
        Ddata: same form as data but containing the uncertainties
        uncert: Maximum uncertainty of the data. Default=20.0
        offset: initial offset. Default=None

    Returns:
        yearf: Time array
        data:  Data
        Ddata: Data uncertainty
        
    """

    # converting to mm
    data *= 1000
    Ddata *= 1000
    return vshift(yearf, data, Ddata, uncert=uncert, offset=offset)


def filt_outl(yearf, data, Ddata, pb, errfunc, outlier):
    """
    This function removes outliers from a time series.

    Examples:
        >>> filt_outl(yearf, data, Ddata, pb, errfunc, outlier)

    Args:
        yearf: time array with numeric time values
        data: data array
        Ddata: same form as data but containing the uncertainties
        pb: parameters of the error function. Default=[None, None, None]
        errfunc: error function. Default=abs
        outlier: Maximum uncertainty of the data. 

    Returns:
        yearf: Time array
        data:  Data
        Ddata: Data uncertainty

    """
    # Removing big outliers
    for i in range(3):
        index = np.where(abs(errfunc(pb[i], yearf - yearf[0], data[i])) > outlier[i])
        yearf = np.delete(yearf, index)
        data = np.delete(data, index, 1)
        Ddata = np.delete(Ddata, index, 1)

    return yearf, data, Ddata


def gamittoNEU(sta, mm=False, ref="plate", dstring=None):
    """
    This function convert a gamit time series to a single np.array with readable time tag.

    Examples:
        >>> gamittoNEU(sta, mm=False, ref="plate", dstring=None)

    Args:
        sta: station name
        mm: convert to mm. Default=False
        ref: reference. Default="plate"
        dstring: time tag. Default=None

    Returns:
        yearf: Time array
        data:  Data
        Ddata: Data uncertainty
        
    """

    import geofunc.geofunc as gf

    yearf, data, Ddata = openGlobkTimes(sta)
    yearf, data, Ddata, _ = vshift(
        yearf, data, Ddata, uncert=1.1, refdate=None, Period=5, offset=None
    )

    # remove plata velocity
    if ref == "plate":
        plateVel = gf.plateVelo([sta])
        data[0, :] = data[0, :] - plateVel[0, 1] * (yearf - yearf[0])
        data[1, :] = data[1, :] - plateVel[0, 0] * (yearf - yearf[0])

    # convert to mm
    if mm:
        data = data * 1000
        Ddata = Ddata * 1000

    return gtoNEU(yearf, data, Ddata, dstring=dstring)


def gtoNEU(yearf, data, Ddata, dstring=None):
    """ 
    This function converts a gamit time series to a single np.array with readable time tag.

    Examples:
        >>> gtoNEU(yearf, data, Ddata)

    Args:
        yearf: time array with numeric time values
        data: data array
        Ddata: same form as data but containing the uncertainties
        dstring: time tag. Default=None

    Returns:
        yearf: Time array
        data:  Data
        Ddata: Data uncertainty

    """
    import numpy as np

    from gtimes.timefunc import convfromYearf

    if dstring == "yearf":  # use the decimal year format
        NEUdata = np.array(
            zip(yearf, data[0], data[1], data[2], Ddata[0], Ddata[1], Ddata[2]),
            dtype=[
                ("yearf", float),
                ("data[0]", float),
                ("data[1]", float),
                ("data[2]", float),
                ("Ddata[0]", float),
                ("Ddata[1]", float),
                ("Ddata[2]", float),
            ],
        )
    else:
        yearf = convfromYearf(yearf, dstring)

        NEUdata = np.array(
            zip(yearf, data[0], data[1], data[2], Ddata[0], Ddata[1], Ddata[2]),
            dtype=[
                ("yearf", "S23"),
                ("data[0]", float),
                ("data[1]", float),
                ("data[2]", float),
                ("Ddata[0]", float),
                ("Ddata[1]", float),
                ("Ddata[2]", float),
            ],
        )

    return NEUdata


def read_gps_data(
    sta,
    Dir=None,
    start=None,
    end=None,
    ref="plate",
    detrend_periodic=True,
    detrend_line=False,
    fit=False,
    detrend_period=[None, None],
    useSTA=None,
    useFIT=None,
    uncert=20.0,
    logging_level=logging.WARNING,
):
    """
    This function reads gps data from a station. It can be used to get the raw data, plate velocity and detrended data.

    Examples:


    Args:
        sta: station name
        Dir: directory. Default=None
        start: start date. Default=None
        end: end date. Default=None
        ref: reference. Default="plate"
        detrend_periodic: detrend periodic. Default=True
        detrend_line: detrend line. Default=False
        fit: fit. Default=False
        detrend_period: detrend period. Default=[None, None]
        useSTA: works in conjunction with useFit to get the fit. Default=None
        useFIT: useFIT from another station. Default=None
        uncert: Maximum uncertainty of the data. Default=20.0
        logging_level: logging level. Default=logging.WARNING
             

    Returns:
        const: detrending constants coefficient, for a lineperiodic function.
        data:  Data array for the three components including the uncertainties from the raw_data (measurement uncertainty)

    """

    import logging
    import numpy as np
    from gtimes.timefunc import TimetoYearf

    # Handling logging
    logging.basicConfig(
        format="%(asctime)-15s [%(levelname)s] %(funcName)s: %(message)s",
        level=logging_level,
    )
    logging.getLogger().setLevel(logging_level)
    module_logger = logging.getLogger()

    yearf, data, Ddata, _ = getData(sta, Dir=Dir, ref=ref)
    syearf, sdata, sDdata = yearf, data, Ddata

    # NOTE: detrending needs to be revised.
    const = getDetrFit(sta, useSTA=useSTA, useFIT=useFIT)
    p0 = convconst(const)
    pb = p0.copy()

    module_logger.info("Fitting constants:\n {}".format(p0))
    if not np.all([p0[i][1:].dtype if p0[i] is not None else False for i in range(3)]):
        module_logger.info("Setting fit to lineperiodic")
        fit = "lineperiodic"

    if useFIT == "periodic":
        module_logger.info(
            '"{}" parameters from {} used in {}'.format(
                const["Fit"].values[0], const["UseSTA"].values[0], sta
            )
        )
        module_logger.info('Setting the fit to a "line" for estimating the rate')
        fit = "line"

    syearf, sdata, sDdata = dPeriod(
        yearf,
        data,
        Ddata,
        startyear=detrend_period[0],
        endyear=detrend_period[1],
    )

    if fit:
        module_logger.info("Fitting =====================")
        if len(syearf) == 0:
            module_logger.warning(
                "No data for interval {}-{} try using the whole time series".format(
                    *detrend_period
                )
            )
            syearf, sdata, sDdata = yearf, data, Ddata

        module_logger.info("Fitting the data to a {}".format(fit))

        if fit == "line":
            pt = [p0[i][0:2] for i in range(3)]
            module_logger.info("initial parameters p0 {}".format(pt))
            pb, _ = fittimes(line, syearf, sdata, sDdata, p0=pt)
            module_logger.info("Estimated parameters for {} {}".format(sta, pb))
            pb = [np.concatenate((pb[i], p0[i][2:])) for i in range(3)]
            pt = [p0[i][0:2] for i in range(3)]

        elif fit == "periodic":
            pb, _ = fittimes(periodic, syearf, sdata, sDdata, p0=p0)

        elif fit == "lineperiodic":
            pb, _ = fittimes(lineperiodic, syearf, sdata, sDdata, p0=p0)

        const = convconst(const, pb)
        const.loc[sta, ["Starttime", "Endtime"]] = [syearf[0], syearf[-1]]

    else:
        module_logger.info("Not fitting")

    if np.all([pb[i][1:] if pb[i] is not None else False for i in range(3)]):
        if detrend_periodic:
            module_logger.info("Periodic detrending")
            module_logger.debug("Fitt parameters: {}".format(pb))
            data = detrend(yearf, data.copy(), Ddata, fitfunc=periodic, p=pb)

        if detrend_line:
            module_logger.info("linear detrending")
            pc = [pb[i][0:2] for i in range(3)]
            module_logger.debug("Fitt parameters: {}".format(pc))
            data = detrend(yearf, data.copy(), Ddata, fitfunc=line, p=pc)
    else:
        module_logger.warn("Fitt parameters are unknown pb={}".format(pb))
        if (detrend_periodic or detrend_line):
            module_logger.warning(
                "Will try to estimate the parameters based the whole dataset and detrend"
            )
            data = detrend(yearf, data.copy(), Ddata, fitfunc=lineperiodic)
    # ----------------------------------------------------

    if start:
        startf = TimetoYearf(start.year, start.month, start.day)
    else:
        startf = yearf[0]
    if end:
        endf = TimetoYearf(end.year, end.month, end.day)
    else:
        endf = yearf[-1]

    yearf, data, Ddata = dPeriod(yearf, data, Ddata, startyear=startf, endyear=endf)
    yearf, data, Ddata, _ = vshift(yearf, data, Ddata, uncert=uncert)

    data = convGlobktopandas(yearf, data, Ddata)

    data["hlength"] = np.sqrt(np.square(data[["east", "north"]]).sum(axis=1))
    data["hangle"] = np.rad2deg(np.arctan2(data["north"], data["east"]))
    data["Dhlength"] = np.sqrt(np.square(data[["Deast", "Dnorth"]]).sum(axis=1))

    module_logger.info("Dataframe columns:\n" + str(data.columns) + "\n")
    module_logger.info("Input time period: ({}, {})".format(start, end))
    module_logger.info("dataframe First and Last lines:\n" + str(data.iloc[[0, -1]]))
    module_logger.info("detrend parameter row:\n" + str(const))
    module_logger.debug("Dataframe shape: {}".format(str(data.shape)))
    module_logger.debug("dataframe types:\n" + str(data.dtypes) + "\n")

    return data, const


#
# Other functions
#


def toDateTime(yearf):
    """
    Function converting from floating point year to datetime.

    Examples:


    Args:
        yearf: floating point year array

    Returns:
        tmp: datetime equivalence of the yearf array



    """
    from gtimes.timefunc import TimefromYearf

    tmp = []

    for i in range(len(yearf)):
        tmp.append(TimefromYearf(yearf[i]))

    return tmp


def toord(yearf):
    """
    Function from floating point year to floating point ordinal.

    Examples:

    Args:
        yearf: floating point year array

    Returns:
        yearf: floating point ordinal array

    """
    from gtimes.timefunc import TimefromYearf

    for i in range(len(yearf)):
        yearf[i] = TimefromYearf(yearf[i], "ordinalf")

    return yearf


def fromord(yearf):
    """
    Function from floating point ordinal to floating point year.

    Examples:

    Args:
        yearf: floating point ordinal array

    Returns:
        yearf: floating point year array

    """


    # from floating point year to floating point ordinal

    from gtimes.timefunc import TimetoYearf

    for i in range(len(yearf)):
        yearf[i] = Timeto(yearf[i], "ordinalf")

    return yearf


def getData(
    sta,
    fstart=None,
    fend=None,
    ref="itrf2008",
    Dir=None,
    tType="TOT",
    uncert=15,
    offset=None,
):
    """
    Function extracting and filtering data to prepeare for plotting.

    Examples:
        >>>getData('GRIC', fstart='2017-01-01', fend='2017-12-31', ref='itrf2008', Dir=None, tType='TOT', uncert=15, offset=None)

    Args:
        sta: station name
        fstart: start date
        fend: end date
        ref: reference frame
        Dir: directory
        tType: type of data
        uncert: uncertainty
        offset: offset

    Returns:
        yearf: floating point year array
        data: data array
        Ddata: uncertainty array
        offset: offset

    


    """

    import geofunc.geofunc as gf

    if tType == "JOIN":
        tType = "TOT"

    yearf, data, Ddata = openGlobkTimes(sta, Dir=Dir, tType=tType)
    yearf, data, Ddata = dPeriod(yearf, data, Ddata, fstart, fend)
    if yearf is None or len(yearf) == 0:
        print("WARNING: no data for station {}".format(sta))
        return None, None, None, None
    yearf, data, Ddata, offset = iprep(yearf, data, Ddata, uncert=uncert, offset=offset)

    if offset is None:
        print("WARNING: offset determination failure for station {}".format(sta))

    if ref == "plate":
        plateVel = gf.plateVelo([sta])
        data[0, :] = data[0, :] - plateVel[0, 1] * 1000 * (yearf - yearf[0])
        data[1, :] = data[1, :] - plateVel[0, 0] * 1000 * (yearf - yearf[0])

    elif ref == "detrend":
        pN, pE, pU, pb = detrend(yearf, data, sta)
        pb_org = [pN, pE, pU]

        for i in range(3):
            data[i] = -errfunc(pb_org[i], yearf - yearf[0], data[i])

    elif ref == "itrf2008":
        pass

    else:
        plateVel = gf.plateVelo([sta], ref)
        data[0, :] = data[0, :] - plateVel[0, 1] * 1000 * (yearf - yearf[0])
        data[1, :] = data[1, :] - plateVel[0, 0] * 1000 * (yearf - yearf[0])

    return yearf, data, Ddata, offset


#
#   --- Private functions ---
#


def __converter(x):
    """
    The data extracted are converted to float and
    ocassional ******* in the data files needs to handled as NAN

    """
    import numpy as np

    try:
        return float(x)
    except:
        return np.nan

    # if x == '********':
    #    return np.nan
    # else:
    #    return float(x)



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
This module contains functions for reading and returning GPS data

The module containes the following functions 
---------------
def read_hytrological(station, start=None, end=None, 
                      frfile=False, fname=None, wrfile=True, flist=[], 
                      base_path="/mnt/hytrodata/", fext=".dat" )
---------------

"""

# used in a a function arguments
import logging

#
# Functions returning data subsets.
#


#
# Functions reading in data from data sources
#


def read_hytrological_data(station, start=None, end=None, frfile=False, fname=None, wrfile=True, 
                           flist=[], base_path="/mnt_data/hytrodata/", fext=".dat", logging_level=logging.WARNING):
        
    """
    ----------------------- ALPHA VERSION --------------------------------
    Reading in hytrological data and returning it
    
    input: 
        station: 


    output:
        return data:
            data is a pandas.DataFrame() containing time series of hytrological data specified in input.
    

    This is the first attempt at general reading in of  hytrological data 
    
    for test and achive purpuses the data can be read in through a file by frfile=True and spesifying a file name
    assuming the data structure was saved using data.to_pickle.

    """
    
    # Imports
    import os
    import logging

    import datetime as dt
    import pandas as pd

    import gtimes.timefunc as gt

    
    tstrf="%Y%m%d %H:%M:%S"
    # defing period to extract
    if end == None:
        end = dt.datetime.now()
    if start == None:
        start = gt.toDatetime("1990-01-01 00:00:00",dstr)
        #start = gt.currDatetime(-90,refday=end) 

    # Handling logging
    logging.basicConfig(
    format="%(asctime)-15s [%(levelname)s] %(funcName)s: %(message)s",
    level=logging_level )
    logging.getLogger().setLevel(logging_level)
    module_logger = logging.getLogger()

    if frfile:
        # For later
        if not fname:
            fname="hytro-data.pik"
        data = pd.read_pickle(fname)

    else:
        
        if flist == []:
            # formatting the file and constructing a list
            var_path="{0}".format(station.capitalize())
            fbasename="{0}_alestur-".format(station.capitalize())
            fstring="{0}%Y{1}".format(os.path.join(base_path, var_path, fbasename),fext)

            # Creating a list of files to open
            flist=gt.datepathlist(fstring,'1Y',start,end, datelist=[ dt.date(year,1,1) for year in range(start.year,end.year+1) ])

            #u_columnlist=["W1_med_cm", "Tw1_med_C", "EC_med_us", "pH_med", "TL1_med_C"]
            # data columns to use
            u_columnlist=["W1_med_cm", "Tw1_med_C", "EC_med_us", "TL1_med_C"]

        data = pd.DataFrame()
        for dfile in flist:
            try:
                module_logger.info("Reading from file: {}".format(dfile))
                tmp = pd.read_csv(dfile,sep=',', header=[0],skiprows=[0,2,3], parse_dates=["TIMESTAMP"] ,index_col=[0], dayfirst=True,
                                  error_bad_lines=False, warn_bad_lines=False, low_memory=False)
                                  # low_memory=False suppresses mixed type warning ???
                module_logger.debug("Dataframe columns: {}:\n".format(dfile) + str(tmp.columns) + "\n")

            except:
                module_logger.warning("File, {0} not found".format(dfile))
                continue
            
            for column in u_columnlist:
                tmp[column] = pd.to_numeric(tmp[column], errors="coerce")
                
            tmp = tmp[u_columnlist].dropna()
            data = data.append(tmp)

        data.drop_duplicates()
        module_logger.info("Dataframe columns:\n" + str(data.columns) + "\n")
        module_logger.info("Input time period: ({}, {})".format(start, end))
        module_logger.info("dataframe First and Last lines:\n"  + str(data.iloc[[0,-1]]))
        module_logger.debug("Dataframe shape: {}".format( str(data.shape) )) 
        module_logger.debug("dataframe types:\n" + str(data.dtypes) + "\n" )


        if wrfile:
            if not fname:
                fname="hytro-data.pik"

            data.to_pickle("{}".format(fname))



    if data.index[0] > start:
        start =  data.index[0]

    if data.index[-1] < end:
        end = data.index[-1]

    return data[start.strftime(tstrf):end.strftime(tstrf)]
